import copy
import http.client
import queue
import socket
import threading
import time
import traceback
import uuid
from http.server import HTTPServer
from urllib.parse import urlparse, urlunparse

from lxml import etree as etree_

from sdc11073.pysoap.soapclient import HTTPReturnCodeError
from sdc11073.pysoap.soapenvelope import Soap12Envelope, ReceivedSoap12Envelope, WsAddress, WsSubscribe, \
    SoapResponseException
from sdc11073.pysoap.soapenvelope import WsaEndpointReferenceType
from .. import commlog
from .. import loghelper
from .. import observableproperties as properties
from .. import xmlparsing, isoduration
from ..httprequesthandler import HTTPRequestHandler
from ..namespaces import Prefix_Namespace as Prefix
from ..namespaces import nsmap as _global_nsmap
from ..namespaces import wseTag, wsaTag, WSA_IS_REFERENCE_PARAMETER

SUBSCRIPTION_CHECK_INTERVAL = 5  # seconds


class MyThreadingMixIn(object):

    def process_request_thread(self, request, client_address):
        """Same as in BaseServer but as a thread.

        In addition, exception handling is done here.

        """
        try:
            self.finish_request(request, client_address)
            self.shutdown_request(request)
        except Exception as ex:
            if self.dispatcher is not None:
                # only
                self.handle_error(request, client_address)
            self.shutdown_request(request)

    def process_request(self, request, client_address):
        """Start a new thread to process the request."""
        t = threading.Thread(target=self.process_request_thread,
                             args=(request, client_address),
                             name='SubscrRecv{}'.format(client_address))
        t.daemon = True
        t.start()
        self.threads.append((t, request, client_address))


class ThreadingHTTPServer(MyThreadingMixIn, HTTPServer):
    """ Each request is handled in a thread.
    Following receipe from https://pymotw.com/2/BaseHTTPServer/index.html#module-BaseHTTPServer
    """

    def __init__(self, *args, **kwargs):
        HTTPServer.__init__(self, *args, **kwargs)
        self.daemon_threads = True
        self.threads = []
        self.dispatcher = None


class ClSubscription(object):
    """ This class handles a subscription to an event source.
    It stores all key data of the subscription and can renew and unsubscribe this subscription."""
    notification = properties.ObservableProperty()
    IDENT_TAG = etree_.QName('http.local.com', 'MyClIdentifier')

    def __init__(self, dpwsHosted, actions, notification_url, endTo_url, ident, xml_validator):
        """
        :param dpwsHosted: the DPWSHosted instance this port type belongs to
        :param actions: list of subscribed actions
        :param notification_url: e.g. http://1.2.3.4:9999, or https://1.2.3.4:9999
        :param endTo_url: url for subscription end message
        :param ident: prefix for logging
        :param xml_validator: allows schema validation
        """
        self.dpwsHosted = dpwsHosted
        self._actions = actions
        self._filter = ' '.join(actions)
        self._notification_url = notification_url
        self.isSubscribed = False
        self._isSubscribedLock = threading.Lock()
        self.end_status = None  # if device sent a SubscriptionEnd message, this contains the status from the message
        self.end_reason = None  # if device sent a SubscriptionEnd message, this contains the reason from the message
        self.expireAt = None
        self.expire_minutes = None
        self.dev_reference_param = None
        self.notifyTo_identifier = etree_.Element(self.IDENT_TAG)
        self.notifyTo_identifier.text = uuid.uuid4().urn

        self._endTo_url = endTo_url
        self._endTo_identifier = etree_.Element(self.IDENT_TAG)
        self._endTo_identifier.text = uuid.uuid4().urn

        self._subscriptionManagerAddress = None
        self._logger = loghelper.getLoggerAdapter('sdc.client.subscr', ident)
        self.eventCounter = 0  # for display purpose, we count notifications
        self.cl_ident = ident
        self._device_epr = urlparse(self.dpwsHosted.endpointReferences[0].address).path
        self._xml_validator = xml_validator

    @property
    def end_to_identifier(self):
        return self._endTo_identifier

    def _mkSubscribeEnvelope(self, subscribe_epr, expire_minutes):
        soapEnvelope = Soap12Envelope(Prefix.partialMap(Prefix.S12, Prefix.WSA, Prefix.WSE))
        soapEnvelope.setAddress(WsAddress(action='http://schemas.xmlsoap.org/ws/2004/08/eventing/Subscribe',
                                          to=subscribe_epr))

        body = WsSubscribe(notifyTo=WsaEndpointReferenceType(self._notification_url,
                                                             referenceParametersNode=[self.notifyTo_identifier]),
                           endTo=WsaEndpointReferenceType(self._endTo_url,
                                                          referenceParametersNode=[self._endTo_identifier]),
                           expires=expire_minutes * 60,
                           filter_=self._filter)
        soapEnvelope.addBodyObject(body)
        return soapEnvelope

    def _handleSubscribeResponse(self, soapEnvelope):
        # Check content of response; raise Error if subscription was not successful
        try:
            msgNode = soapEnvelope.msgNode
            if msgNode.tag == wseTag('SubscribeResponse'):
                address = msgNode.xpath('wse:SubscriptionManager/wsa:Address/text()', namespaces=_global_nsmap)
                self.dev_reference_param = None

                reference_params = msgNode.xpath('wse:SubscriptionManager/wsa:ReferenceParameters',
                                                 namespaces=_global_nsmap)
                if reference_params:
                    self.dev_reference_param = reference_params[0]
                expires = msgNode.xpath('wse:Expires/text()', namespaces=_global_nsmap)

                self._subscriptionManagerAddress = urlparse(address[0])
                expireseconds = isoduration.parse_duration(expires[0])
                self.expireAt = time.time() + expireseconds
                self.isSubscribed = True
                self._logger.info('Subscribe was successful: expires at {}, address="{}"',
                                  self.expireAt, self._subscriptionManagerAddress)
            else:
                # This is a failure response or even rubbish. log it and raise error
                self._logger.error('Subscribe response has unexpected content: {}', soapEnvelope.as_xml(pretty=True))
                self.isSubscribed = False
                raise SoapResponseException(soapEnvelope)
        except AttributeError:
            self._logger.error('Subscribe response has unexpected content: {}', soapEnvelope.as_xml(pretty=True))
            self.isSubscribed = False
            raise SoapResponseException(soapEnvelope)

    def subscribe(self, expire_minutes=60):
        self._logger.info('### startSubscription "{}" ###', self._filter)
        self.eventCounter = 0
        self.expire_minutes = expire_minutes  # saved for later renewal, we will use the same interval
        # ToDo: check if there is more than one address. In that case a clever selection is needed
        address = self.dpwsHosted.endpointReferences[0].address
        soapEnvelope = self._mkSubscribeEnvelope(address, expire_minutes)
        msg = 'subscribe {}'.format(self._filter)
        try:
            resultSoapEnvelope = self.dpwsHosted.soapClient.postSoapEnvelopeTo(self._device_epr, soapEnvelope, msg=msg)
            self._handleSubscribeResponse(resultSoapEnvelope)
        except HTTPReturnCodeError as ex:
            self._logger.error('could not subscribe: {}'.format(HTTPReturnCodeError))

    def _add_device_references(self, soapEnvelope):
        """ add references for requests to device (renew, getstatus, unsubscribe)"""
        if self.dev_reference_param is not None:
            for e in self.dev_reference_param:
                e_ = copy.copy(e)
                # mandatory attribute acc. to ws_addressing SOAP Binding (https://www.w3.org/TR/2006/REC-ws-addr-soap-20060509/)
                e_.set(wsaTag('IsReferenceParameter'), 'true')
                soapEnvelope.addHeaderElement(e_)

    def _mkRenewEnvelope(self, expire_minutes):
        soapEnvelope = Soap12Envelope(Prefix.partialMap(Prefix.S12, Prefix.WSA, Prefix.WSE))
        soapEnvelope.setAddress(WsAddress(action='http://schemas.xmlsoap.org/ws/2004/08/eventing/Renew',
                                          to=urlunparse(self._subscriptionManagerAddress)))
        self._add_device_references(soapEnvelope)
        renewNode = etree_.Element(wseTag('Renew'), nsmap=Prefix.partialMap(Prefix.WSE))
        expiresNode = etree_.SubElement(renewNode, wseTag('Expires'), nsmap=Prefix.partialMap(Prefix.WSE))
        expiresNode.text = isoduration.durationString(expire_minutes * 60)
        soapEnvelope.addBodyElement(renewNode)
        return soapEnvelope

    def _handleRenewResponse(self, soapEnvelope):
        # Check content of response; raise Error if subscription was not successful
        bodyNode = soapEnvelope.bodyNode
        renewResponse = bodyNode.xpath('wse:RenewResponse', namespaces=_global_nsmap)
        if len(renewResponse) == 1:
            # this means renew was accepted
            expires = bodyNode.xpath('wse:RenewResponse/wse:Expires/text()', namespaces=_global_nsmap)
            expireseconds = isoduration.parse_duration(expires[0])
            self.expireAt = time.time() + expireseconds
        else:
            raise SoapResponseException(soapEnvelope)

    def renew(self, expire_minutes=60):
        with self._isSubscribedLock:
            if not self.isSubscribed:
                return None

            soapEnvelope = self._mkRenewEnvelope(expire_minutes)
            try:
                resultSoapEnvelope = self.dpwsHosted.soapClient.postSoapEnvelopeTo(self._subscriptionManagerAddress.path,
                                                                                   soapEnvelope, msg='renew')
                self._logger.debug('{}', resultSoapEnvelope.as_xml(pretty=True))
            except HTTPReturnCodeError as ex:
                self.isSubscribed = False
                self._logger.error('could not renew: {}'.format(HTTPReturnCodeError))
            except (http.client.HTTPException, ConnectionError) as ex:
                self._logger.warn('renew failed: {}', ex)
                self.isSubscribed = False
            except Exception as ex:
                self._logger.error('Exception in renew: {}', ex)
                self.isSubscribed = False
            else:
                try:
                    self._handleRenewResponse(resultSoapEnvelope)
                    return self.remainingSubscriptionSeconds
                except SoapResponseException as ex:
                    self.isSubscribed = False
                    self._logger.warn('renew failed: {}',
                                      etree_.tostring(ex.soapResponseEnvelope.bodyNode, pretty_print=True))

    def unsubscribe(self):
        with self._isSubscribedLock:
            if not self.isSubscribed:
                return
            self.isSubscribed = False  # handle it as unsubscribed from now on
            soapEnvelope = Soap12Envelope(Prefix.partialMap(Prefix.S12, Prefix.WSA, Prefix.WSE))
            soapEnvelope.setAddress(WsAddress(action='http://schemas.xmlsoap.org/ws/2004/08/eventing/Unsubscribe',
                                              to=urlunparse(self._subscriptionManagerAddress)))
            self._add_device_references(soapEnvelope)
            soapEnvelope.addBodyElement(etree_.Element(wseTag('Unsubscribe')))
            resultSoapEnvelope = self.dpwsHosted.soapClient.postSoapEnvelopeTo(self._subscriptionManagerAddress.path,
                                                                               soapEnvelope, msg='unsubscribe')
            responseAction = resultSoapEnvelope.address.action
            # check response: response does not contain explicit status. If action== UnsubscribeResponse all is fine.
            if responseAction == 'http://schemas.xmlsoap.org/ws/2004/08/eventing/UnsubscribeResponse':
                self._logger.info('unsubscribe: end of subscription {} was confirmed.', self._filter)
            else:
                self._logger.error('unsubscribe: unexpected response action: {}', resultSoapEnvelope.as_xml(pretty=True))
                raise RuntimeError(
                    'unsubscribe: unexpected response action: {}'.format(resultSoapEnvelope.as_xml(pretty=True)))

    def _mkGetStatusEnvelope(self):
        soapEnvelope = Soap12Envelope(Prefix.partialMap(Prefix.S12, Prefix.WSA, Prefix.WSE))
        soapEnvelope.setAddress(WsAddress(action='http://schemas.xmlsoap.org/ws/2004/08/eventing/GetStatus',
                                          to=urlunparse(self._subscriptionManagerAddress)))
        self._add_device_references(soapEnvelope)
        bodyNode = etree_.Element(wseTag('GetStatus'))
        soapEnvelope.addBodyElement(bodyNode)
        return soapEnvelope

    def getStatus(self):
        """ Sends a GetStatus Request to the device.
        :return: the remaining time of the subscription or None, if the request was not successful
        """
        with self._isSubscribedLock:
            if not self.isSubscribed:
                return None
            soapEnvelope = self._mkGetStatusEnvelope()
            try:
                resultSoapEnvelope = self.dpwsHosted.soapClient.postSoapEnvelopeTo(self._subscriptionManagerAddress.path,
                                                                                   soapEnvelope, msg='getStatus')
            except HTTPReturnCodeError as ex:
                self.isSubscribed = False
                self._logger.error('could not get status: {}'.format(HTTPReturnCodeError))
            except (http.client.HTTPException, ConnectionError) as ex:
                self.isSubscribed = False
                self._logger.warn('getStatus: Connection Error {} for subscription {}', ex, self._filter)
            except Exception as ex:
                self._logger.error('Exception in getStatus: {}', ex)
                self.isSubscribed = False
            else:
                try:
                    expiresNode = resultSoapEnvelope.msgNode.find('wse:Expires', namespaces=_global_nsmap)
                    if expiresNode is None:
                        self._logger.warn('getStatus for {}: Could not find "Expires" node! getStatus={} ', self._filter,
                                          resultSoapEnvelope.rawdata)
                        raise SoapResponseException(resultSoapEnvelope)
                    else:
                        expires = expiresNode.text
                        expiresValue = isoduration.parse_duration(expires)
                        self._logger.debug('getStatus for {}: Expires = {} = {} seconds, counter = {}', self._filter,
                                           expires,
                                           expiresValue,
                                           self.eventCounter)
                        return expiresValue
                except AttributeError:
                    self._logger.warn('No msg in envelope')

    def checkStatus(self, renewLimit):
        """ Calls getStatus and updates internal data.
        :param renewLimit: a value in seconds. If remaining duration of subscription is less than this value, it renews the subscription.
        :return: None
        """
        remainingTime = self.getStatus()
        if not self.isSubscribed:
            return
        elif abs(remainingTime - self.remainingSubscriptionSeconds) > 10:
            self._logger.warn(
                'time delta between expected expire and reported expire  > 10 seconds. Will correct own expectation.')
            self.expireAt = time.time() + remainingTime

        if self.remainingSubscriptionSeconds < renewLimit:
            self._logger.info('renewing subscription')
            self.renew()

    @property
    def remainingSubscriptionSeconds(self):
        return self.expireAt - time.time()

    def onNotification(self, soapEnvelope):
        try:
            soapEnvelope.validate_envelope(self._xml_validator)
        except etree_.DocumentInvalid as ex:
            self._logger.error('received invalid document: {}', ex)
        except Exception as ex:
            self._logger.error('error validation document: {}', ex)
        self.eventCounter += 1
        self.notification = soapEnvelope

    @property
    def shortFilterString(self):
        return xmlparsing.shortFilterString(self._actions)

    def __str__(self):
        return 'Subscription of "{}", isSubscribed={}, remaining time = {} sec., count={}'.format(
            self.shortFilterString,
            self.isSubscribed,
            int(self.remainingSubscriptionSeconds),
            self.eventCounter)


class SubscriptionManager(threading.Thread):
    """ Factory for Subscription objects, thread that automatically renews expiring subscriptions.
     """
    allSubscriptionsOkay = properties.ObservableProperty(True)  # a boolean
    keepAlive_with_renew = True  # enable as workaround if checkstatus is not supported

    def __init__(self, notification_url, endTo_url=None, checkInterval=None, log_prefix='', xml_validator=None):
        """
        :param notification_url: the destination url for notifications.
        :param endTo_url: if given the destination url for end subscription notifications; if not given, the notification_url is used.
        :param checkInterval: the interval (in seconds ) for getStatus requests. Defaults to SUBSCRIPTION_CHECK_INTERVAL
        :param log_prefix: a string that is used in log output; defaults to empty string
        :param xml_validator: helper for schema validation
         """
        super().__init__(name='Cl_SubscriptionManager{}'.format(log_prefix))
        self.daemon = True
        self._checkInterval = checkInterval or SUBSCRIPTION_CHECK_INTERVAL
        self.subscriptions = {}
        self._subscriptionsLock = threading.Lock()

        self._run = False
        self._notification_url = notification_url
        self._endTo_url = endTo_url or notification_url
        self._logger = loghelper.getLoggerAdapter('sdc.client.subscrMgr', log_prefix)
        self.log_prefix = log_prefix
        self._xml_validator = xml_validator

    def stop(self):
        self._run = False
        self.join(timeout=2)
        with self._subscriptionsLock:
            self.subscriptions.clear()

    def run(self):
        self._run = True
        try:
            while self._run:
                try:
                    for i in range(self._checkInterval):
                        time.sleep(1)
                        if not self._run:
                            return
                        # check if all subscriptions are okay
                        with self._subscriptionsLock:
                            not_okay = [s for s in self.subscriptions.values() if not s.isSubscribed]
                            subscribed = [s.isSubscribed for s in self.subscriptions.values()]
                            self.allSubscriptionsOkay = (len(not_okay) == 0)
                    with self._subscriptionsLock:
                        subscriptions = list(self.subscriptions.values())
                    for subscription in subscriptions:
                        if self.keepAlive_with_renew:
                            subscription.renew()
                        else:
                            subscription.checkStatus(renewLimit=self._checkInterval * 5)
                    self._logger.debug('##### SubscriptionManager Interval ######')
                    for subscription in subscriptions:
                        self._logger.debug('{}', subscription)
                except Exception as ex:
                    self._logger.error('##### check loop: {}', traceback.format_exc())
        finally:
            self._logger.info('terminating subscriptions check loop! self._run={}', self._run)

    def mkSubscription(self, dpwsHosted, filters):
        s = ClSubscription(dpwsHosted, filters, self._notification_url, self._endTo_url, self.log_prefix,
                           self._xml_validator)
        filter_ = ' '.join(filters)
        with self._subscriptionsLock:
            self.subscriptions[filter_] = s
        return s

    def onSubScriptionEnd(self, soapenvelope):
        subscr_ident_list = soapenvelope.headerNode.findall(ClSubscription.IDENT_TAG, namespaces=_global_nsmap)
        statuus = soapenvelope.bodyNode.xpath('wse:SubscriptionEnd/wse:Status/text()', namespaces=_global_nsmap)
        reasons = soapenvelope.bodyNode.xpath('wse:SubscriptionEnd/wse:Reason/text()', namespaces=_global_nsmap)
        if statuus:
            info = ' status={} '.format(statuus[0])
        else:
            info = ''
        if reasons:
            if len(reasons) == 1:
                info += ' reason = {}'.format(reasons[0])
            else:
                info += ' reasons = {}'.format(reasons)
        if not subscr_ident_list:
            self._logger.warn('onSubScriptionEnd: did not find any identifier in message')
            return None
        subscr_ident = subscr_ident_list[0]
        for s in self.subscriptions.values():
            if subscr_ident.text == s.end_to_identifier.text:
                self._logger.info('onSubScriptionEnd: received Subscription End for {} {}',
                                  s.shortFilterString,
                                  info)
                s.isSubscribed = False
                if len(statuus) > 0:
                    s.end_status = statuus[0]
                if len(reasons) > 0:
                    s.end_reason = reasons[0]
                return s
        self._logger.warn('onSubScriptionEnd: have no subscription for identifier = {}', subscr_ident.text)
        return None

    def unsubscribeAll(self):
        with self._subscriptionsLock:
            current_subscriptions = list(self.subscriptions.values())  # make a copy
            self.subscriptions.clear()
            for s in current_subscriptions:
                try:
                    s.unsubscribe()
                except:
                    self._logger.warn('unsubscribe error: {}\n call stack:{} ', traceback.format_exc(),
                                      traceback.format_stack())


class _DispatchError(Exception):
    def __init__(self, httpErrorcode, errorText):
        super(_DispatchError, self).__init__()
        self.httpErrorcode = httpErrorcode
        self.errorText = errorText


class SOAPNotificationsDispatcher(object):
    """ receiver of all notifications"""

    def __init__(self, log_prefix, sdc_definitions):
        self._logger = loghelper.getLoggerAdapter('sdc.client.notif_dispatch', log_prefix)
        self.log_prefix = log_prefix
        self._sdc_definitions = sdc_definitions
        self.methods = {}

    def register_function(self, action, ref, fn):
        self.methods[(ref, action)] = fn

    def dispatch(self, path, xml):
        start = time.time()

        fn, action, request = self._dispatch_process_data(xml, path)
        fn(request)

        duration = time.time() - start
        if duration > 0.005:
            self._logger.debug('action {}: duration = {:.4f}sec', action, duration)
        return ''

    def _dispatch_process_data(self, xml, path):
        normalized_xml = self._sdc_definitions.normalizeXMLText(xml)
        request = ReceivedSoap12Envelope.fromXMLString(normalized_xml)
        try:
            action = request.address.action
        except AttributeError:
            raise _DispatchError(404, 'no action in request')
        self._logger.debug('received notification path={}, action = {}', path, action)

        ref_parameters = request.headerNode.findall(ClSubscription.IDENT_TAG, namespaces=_global_nsmap)
        ref_parameters = [i for i in ref_parameters if i.get(WSA_IS_REFERENCE_PARAMETER) in ("1", "true")]

        if len(ref_parameters) != 1:
            raise _DispatchError(400, f'Expected exactly one matching reference parameter missing from request.')

        try:
            fn = self.methods[(ref_parameters[0].text, action)]
        except KeyError:
            self._logger.error('action "{}" not registered. Known:{}'.format(action, self.methods.keys()))
            raise _DispatchError(404, 'action not registered')

        return fn, action, request


class SOAPNotificationsDispatcherThreaded(SOAPNotificationsDispatcher):

    def __init__(self, ident, bicepsSchema):
        super(SOAPNotificationsDispatcherThreaded, self).__init__(ident, bicepsSchema)
        self._queue = queue.Queue(1000)
        self._worker = threading.Thread(target=self._readqueue)
        self._worker.daemon = True
        self._worker.start()

    def dispatch(self, path, xml):
        fn, action, request = self._dispatch_process_data(xml, path)
        self._queue.put((fn, request, action))
        return ''

    def _readqueue(self):
        while True:
            fn, request, action = self._queue.get()
            try:
                fn(request)
            except:
                self._logger.error(
                    'method {} for action "{}" failed:{}'.format(fn.__name__, action, traceback.format_exc()))


class SOAPNotificationsHandler(HTTPRequestHandler):
    disable_nagle_algorithm = True
    wbufsize = 0xffff  # 64k buffer to prevent tiny packages
    RESPONSE_COMPRESS_MINSIZE = 256  # bytes, compress response it it is larger than this value (and other side supports compression)

    def do_POST(self):
        """SOAP POST gateway"""
        self.server.threadObj._logger.debug('notification do_POST incoming')  # pylint: disable=protected-access
        dispatcher = self.server.dispatcher
        response_string = ''
        if dispatcher is None:
            # close this connection
            self.close_connection = 1
            self.server.threadObj._logger.warn(
                'received a POST request, but no dispatcher => returning 404 ')  # pylint:disable=protected-access
            self.send_response(404)  # not found
        else:
            request_bytes = self._read_request()

            self.server.threadObj._logger.debug('notification {} bytes',
                                                request_bytes)  # pylint: disable=protected-access
            # execute the method
            commlog.defaultLogger.logSoapSubscrMsgIn(request_bytes)
            try:
                response_string = self.server.dispatcher.dispatch(self.path, request_bytes)
                if response_string is None:
                    response_string = ''
                self.send_response(202, b'Accepted')
            except _DispatchError as ex:
                self.server.threadObj._logger.error('received a POST request, but got _DispatchError => returning {}',
                                                    ex.httpErrorcode)  # pylint:disable=protected-access
                self.send_response(ex.httpErrorcode, ex.errorText)
            except Exception as ex:
                self.server.threadObj._logger.error(
                    'received a POST request, but got Exception "{}"=> returning {}\n{}', ex, 500,
                    traceback.format_exc())  # pylint:disable=protected-access
                self.send_response(500, b'server error in dispatch')
        response_bytes = response_string.encode('utf-8')
        if len(response_bytes) > self.RESPONSE_COMPRESS_MINSIZE:
            response_bytes = self._compressIfRequired(response_bytes)

        self.send_header("Content-Type", "application/soap+xml; charset=utf-8")
        self.send_header("Content-Length", len(response_bytes))  # this is necessary for correct keep-alive handling!
        self.end_headers()
        self.wfile.write(response_bytes)


class NotificationsReceiverDispatcherThread(threading.Thread):

    def __init__(self, my_ipaddress, sslContext, log_prefix, sdc_definitions, supportedEncodings,
                 soap_notifications_handler_class=None, async_dispatch=True):
        """

        :param my_ipaddress: http server will listen on this address
        :param sslContext: http server uses this ssl context
        :param log_prefix: used for logging
        :param sdc_definitions: namespaces etc.
        :param supportedEncodings: a list of strings
        :param soap_notifications_handler_class: if None, SOAPNotificationsHandler is used,
                otherwise the provided class ( a HTTPRequestHandler).
        :param async_dispatch: if True, incoming requests are queued and response is sent (processing is done later).
                                if False, response is sent after the complete processing is done.
        """
        super(NotificationsReceiverDispatcherThread, self).__init__(
            name='Cl_NotificationsReceiver_{}'.format(log_prefix))
        self._sslContext = sslContext
        self._soap_notifications_handler_class = soap_notifications_handler_class
        self.daemon = True
        self._logger = loghelper.getLoggerAdapter('sdc.client.notif_dispatch', log_prefix)

        self._my_ipaddress = my_ipaddress
        self.my_port = None
        self.base_url = None
        self.httpd = None
        self.supportedEncodings = supportedEncodings
        # create and set up the dispatcher for notifications
        if async_dispatch:
            self.dispatcher = SOAPNotificationsDispatcherThreaded(log_prefix, sdc_definitions)
        else:
            self.dispatcher = SOAPNotificationsDispatcher(log_prefix, sdc_definitions)
        self.started_evt = threading.Event()  # helps to wait until thread has initialised is variables

    def run(self):
        try:
            myport = 0  # zero means that OS selects a free port
            self.httpd = ThreadingHTTPServer((self._my_ipaddress, myport),
                                             self._soap_notifications_handler_class or SOAPNotificationsHandler)
            # add use compression flag to the server
            setattr(self.httpd, 'supportedEncodings', self.supportedEncodings)
            self.my_port = self.httpd.server_port
            self._logger.info('starting Notification receiver on {}:{}', self._my_ipaddress, self.my_port)
            if self._sslContext:
                self.httpd.socket = self._sslContext.wrap_socket(self.httpd.socket,
                                                                 server_side=True,
                                                                 do_handshake_on_connect=False)
                self.base_url = 'https://{}:{}/'.format(self._my_ipaddress, self.my_port)
            else:
                self.base_url = 'http://{}:{}/'.format(self._my_ipaddress, self.my_port)
            self.httpd.dispatcher = self.dispatcher
            self.httpd.threadObj = self  # make logger available for SOAPNotificationsHandler
            self.started_evt.set()
            self.httpd.serve_forever()
        except Exception:
            self._logger.error(
                'Unhandled Exception at thread runtime. Thread will abort! {}'.format(traceback.format_exc()))
            raise

    def stop(self, closeAllConnections=True):
        """
        :param closeAllConnections: for testing purpose one might want to keep the connection handler threads alive.
                If param is False then they are kept alive.
        """
        self.httpd.shutdown()
        self.httpd.socket.close()
        if closeAllConnections:
            if self.httpd.dispatcher is not None:
                self.httpd.dispatcher.methods = {}
                self.httpd.dispatcher = None  # this leads to a '503' reaction in SOAPNotificationsHandler
            for thr in self.httpd.threads:
                thread, request, client_addr = thr
                if thread.is_alive():
                    try:
                        request.shutdown(socket.SHUT_RDWR)
                        request.close()
                        self._logger.info('closed socket for notifications from {}', client_addr)
                    except OSError as ex:
                        # the connection is already closed
                        continue
                    except Exception as ex:
                        self._logger.warn('error closing socket for notifications from {}: {}', client_addr, ex)
            time.sleep(0.1)
            for thr in self.httpd.threads:
                thread, request, client_addr = thr
                if thread.is_alive():
                    thread.join(2)
            del self.httpd.threads[:]



import time
from lxml import etree as etree_
from .. import namespaces
from ..pmtypes import ContextAssociation
from..mdib import msgreader
from . import providerbase


class GenericContextProvider(providerbase.ProviderRole):
    """ Handles SetContextState operations"""
    def __init__(self, opTargetDescriptorTypes=None, forcedNewStateType=None, log_prefix=None):
        super(GenericContextProvider, self).__init__(log_prefix)
        self._opTargetDescrTypes = opTargetDescriptorTypes
        self._forcedNewStateType = forcedNewStateType

    def makeOperationInstance(self, operationDescriptorContainer):
        """Create a handler for SetContextStateOperationDescriptor if type of operation target
        matches opTargetDescriptorTypes"""
        if operationDescriptorContainer.NODETYPE == namespaces.domTag('SetContextStateOperationDescriptor'):
            opTargetDescrContainer = self._mdib.descriptions.handle.getOne(operationDescriptorContainer.OperationTarget)
            if (not self._opTargetDescrTypes) or (opTargetDescrContainer.NODETYPE not in self._opTargetDescrTypes):
                return # we do not handle this target type
            else:
                pc_operation = self._mkOperationFromOperationDescriptor(operationDescriptorContainer,
                                                                        currentArgumentHandler=self._setContextState)
                return pc_operation

    def _setContextState(self, operationInstance, proposedContextStates):
        """ This is the code that executes the operation itself.
        """
        with self._mdib.mdibUpdateTransaction() as tr:
            for proposed_st in proposedContextStates:
                oldContextStateContainer = None
                if proposed_st.descriptorHandle != proposed_st.Handle:
                    # this is an update for an existing state
                    oldContextStateContainer = operationInstance.operationTargetStorage.handle.getOne(
                        proposed_st.Handle, allowNone=True)
                    if oldContextStateContainer is None:
                        raise ValueError('handle {} not found'.format(proposed_st.Handle))
                if oldContextStateContainer is None:
                    # this is a new context state
                    # create a new unique handle
                    handleString = '{}_{}'.format(proposed_st.descriptorHandle, self._mdib.mdibVersion)
                    proposed_st.Handle = handleString
                    if proposed_st.ContextAssociation == ContextAssociation.ASSOCIATED:
                        proposed_st.BindingMdibVersion = self._mdib.mdibVersion + 1
                        proposed_st.BindingStartTime = time.time()
                    proposed_st.updateNode()
                    self._logger.info('new {}, handle={}', proposed_st.NODETYPE.localname, proposed_st.Handle)
                    tr.addContextState(proposed_st)

                    if proposed_st.NODETYPE in (namespaces.domTag('PatientContextState'),
                                                namespaces.domTag('LocationContextState')) \
                            and proposed_st.ContextAssociation == ContextAssociation.ASSOCIATED:
                        # find all associated context states, disassociate them, set unbinding info, and add them to updates
                        oldContextStateContainers = operationInstance.operationTargetStorage.descriptorHandle.get(proposed_st.descriptorHandle, [])
                        for old_st in oldContextStateContainers:
                            if old_st.ContextAssociation != ContextAssociation.DISASSOCIATED or old_st.UnbindingMdibVersion is None:
                                new_st = tr.getContextState(old_st.descriptorHandle, old_st.Handle)
                                new_st.ContextAssociation = ContextAssociation.DISASSOCIATED
                                if new_st.UnbindingMdibVersion is None:
                                    new_st.UnbindingMdibVersion = self._mdib.mdibVersion + 1
                                    new_st.BindingEndTime = time.time()
                else:
                    # this is an update to an existing patient
                    # use "regular" way to update via transaction manager
                    self._logger.info('update {}, handle={}', proposed_st.NODETYPE.localname, proposed_st.Handle)
                    tmp = tr.getContextState(proposed_st.descriptorHandle,
                                             contextStateHandle=proposed_st.Handle)
                    tmp.updateFromOtherContainer(proposed_st, skippedProperties=['ContextAssociation',
                                                                                 'BindingMdibVersion',
                                                                                 'UnbindingMdibVersion',
                                                                                 'BindingStartTime',
                                                                                 'BindingEndTime',
                                                                                 'StateVersion'])


class EnsembleContextProvider(GenericContextProvider):
    def __init__(self, log_prefix):
        super(EnsembleContextProvider, self).__init__(opTargetDescriptorTypes=[namespaces.domTag('EnsembleContextDescriptor')], log_prefix=log_prefix)


class LocationContextProvider(GenericContextProvider):
    def __init__(self, log_prefix):
        super(LocationContextProvider, self).__init__(opTargetDescriptorTypes=[namespaces.domTag('LocationContextDescriptor')], log_prefix=log_prefix)



import os
from lxml import etree as etree_
from .namespaces import Prefix_Namespace as Prefix
#from .definitions_base import SchemaResolverBase
from .definitions_base import BaseDefinitions
from .mdib import descriptorcontainers as dc_final
from .mdib import statecontainers as sc_final
from .schema_resolver import SchemaResolver, mk_schema_validator
schemaFolder = os.path.join(os.path.dirname(__file__), 'xsd')

# the following namespace definitions reflect the initial SDC standard.
# There might be changes or additions in the future, who knows...
#

_DPWS_SDCNamespace = 'http://standards.ieee.org/downloads/11073/11073-20701-2018'
_ActionsNamespace = _DPWS_SDCNamespace

class _SDC_v1_Actions(object):
    OperationInvokedReport = _ActionsNamespace + '/SetService/OperationInvokedReport'
    SetOperationInvokedReport = _ActionsNamespace + '/SetService/OperationInvokedReport'
    ContextOperationInvokedReport =  _ActionsNamespace + '/ContextService/OperationInvokedReport'
    EpisodicContextReport = _ActionsNamespace + '/ContextService/EpisodicContextReport'
    EpisodicMetricReport = _ActionsNamespace + '/StateEventService/EpisodicMetricReport'
    EpisodicOperationalStateReport = _ActionsNamespace + '/StateEventService/EpisodicOperationalStateReport'
    EpisodicAlertReport = _ActionsNamespace + '/StateEventService/EpisodicAlertReport'
    EpisodicComponentReport = _ActionsNamespace + '/StateEventService/EpisodicComponentReport'
    PeriodicContextReport = _ActionsNamespace + '/ContextService/PeriodicContextReport'
    PeriodicMetricReport = _ActionsNamespace + '/StateEventService/PeriodicMetricReport'
    PeriodicOperationalStateReport = _ActionsNamespace + '/StateEventService/PeriodicOperationalStateReport'
    PeriodicAlertReport = _ActionsNamespace + '/StateEventService/PeriodicAlertReport'
    PeriodicComponentReport = _ActionsNamespace + '/StateEventService/PeriodicComponentReport'
    SystemErrorReport = _ActionsNamespace + '/StateEventService/SystemErrorReport'
    Waveform = _ActionsNamespace + '/WaveformService/WaveformStream'
    DescriptionModificationReport = _ActionsNamespace + '/DescriptionEventService/DescriptionModificationReport'
    GetMdib = _ActionsNamespace + '/GetService/GetMdib'
    GetMdibResponse = _ActionsNamespace + '/GetService/GetMdibResponse'
    GetMdState = _ActionsNamespace + '/GetService/GetMdState'
    GetMdStateResponse = _ActionsNamespace + '/GetService/GetMdStateResponse'
    GetMdDescription = _ActionsNamespace + '/GetService/GetMdDescription'
    GetMdDescriptionResponse = _ActionsNamespace + '/GetService/GetMdDescriptionResponse'
    GetContainmentTree = _ActionsNamespace + '/GetService/GetContainmentTree'
    GetContainmentTreeResponse = _ActionsNamespace + '/GetService/GetContainmentTreeResponse'
    GetContextStates = _ActionsNamespace + '/ContextService/GetContextStates'
    GetContextStatesResponse = _ActionsNamespace + '/ContextService/GetContextStatesResponse'
    GetContextStatesByIdentification = _ActionsNamespace + '/ContextService/GetContextStatesByIdentification'
    GetContextStatesByIdentificationResponse = _ActionsNamespace + '/ContextService/GetContextStatesByIdentificationResponse'
    SetContextState = _ActionsNamespace + '/ContextService/SetContextState'
    SetContextStateResponse = _ActionsNamespace + '/ContextService/SetContextStateResponse'
    GetSupportedLanguages = _ActionsNamespace + '/LocalizationService/GetSupportedLanguages'
    GetSupportedLanguagesResponse = _ActionsNamespace + '/LocalizationService/GetSupportedLanguagesResponse'
    GetLocalizedText = _ActionsNamespace + '/LocalizationService/GetLocalizedText'
    GetLocalizedTextResponse = _ActionsNamespace + '/LocalizationService/GetLocalizedTextResponse'
    Activate = _ActionsNamespace + '/SetService/Activate'
    ActivateResponse = _ActionsNamespace + '/SetService/ActivateResponse'
    SetString = _ActionsNamespace + '/SetService/SetString'
    SetStringResponse = _ActionsNamespace + '/SetService/SetStringResponse'
    SetValue = _ActionsNamespace + '/SetService/SetValue'
    SetValueResponse = _ActionsNamespace + '/SetService/SetValueResponse'
    SetAlertState = _ActionsNamespace + '/SetService/SetAlertState'
    SetAlertStateResponse = _ActionsNamespace + '/SetService/SetAlertStateResponse'
    SetMetricState = _ActionsNamespace + '/SetService/SetMetricState'
    SetMetricStateResponse = _ActionsNamespace + '/SetService/SetMetricStateResponse'
    SetComponentState = _ActionsNamespace + '/SetService/SetComponentState'
    SetComponentStateResponse = _ActionsNamespace + '/SetService/SetComponentStateResponse'
    GetDescriptor = _ActionsNamespace + '/ContainmentTreeService/GetDescriptor'
    GetDescriptorResponse = _ActionsNamespace + '/ContainmentTreeService/GetDescriptorResponse'
    GetContainmentTree = _ActionsNamespace + '/ContainmentTreeService/GetContainmentTree'
    GetContainmentTreeResponse = _ActionsNamespace + '/ContainmentTreeService/GetContainmentTreeResponse'
    SubscriptionEnd = Prefix.WSE.namespace + '/SubscriptionEnd'


class SchemaPathsSdc:
    MetaDataExchangeSchemaFile = os.path.join(schemaFolder, 'MetadataExchange.xsd')
    EventingSchemaFile = os.path.join(schemaFolder, 'eventing.xsd')
    SoapEnvelopeSchemaFile = os.path.join(schemaFolder, 'soap-envelope.xsd')
    WsAddrSchemaFile = os.path.join(schemaFolder, 'ws-addr.xsd')
    XMLSchemaFile = os.path.join(schemaFolder, 'xml.xsd')
    DPWSSchemaFile = os.path.join(schemaFolder, 'wsdd-dpws-1.1-schema-os.xsd')
    WSDiscoverySchemaFile = os.path.join(schemaFolder, 'wsdd-discovery-1.1-schema-os.xsd')
    WSDLSchemaFile = os.path.join(schemaFolder, 'wsdl.xsd')
    MessageModelSchemaFile = os.path.join(schemaFolder, 'BICEPS_MessageModel.xsd')
    ParticipantModelSchemaFile = os.path.join(schemaFolder, 'BICEPS_ParticipantModel.xsd')
    ExtensionPointSchemaFile = os.path.join(schemaFolder, 'ExtensionPoint.xsd')
    schema_location_lookup = {  # for schema resolver
        # eventing.xsd originally uses http://schemas.xmlsoap.org/ws/2004/08/addressing,
        # but DPWS overwrites the addressing standard of eventing with http://www.w3.org/2005/08/addressing!
        # In order to reflect this, eventing.xsd was patched so that it uses same namespace and schema location
        # dpws uses a schema location that does not match the namespace!
        'http://www.w3.org/2006/03/addressing/ws-addr.xsd': WsAddrSchemaFile,  # schema loc. used by dpws 1.1 schema
        'http://www.w3.org/2005/08/addressing': WsAddrSchemaFile,
        'http://www.w3.org/2001/xml.xsd': XMLSchemaFile,
        'http://www.w3.org/2003/05/soap-envelope': SoapEnvelopeSchemaFile,
        'http://standards.ieee.org/downloads/11073/11073-10207-2017/BICEPS_ParticipantModel.xsd': ParticipantModelSchemaFile,
        'http://standards.ieee.org/downloads/11073/11073-10207-2017/BICEPS_MessageModel.xsd': MessageModelSchemaFile,
        'http://standards.ieee.org/downloads/11073/11073-10207-2017/ExtensionPoint.xsd': ExtensionPointSchemaFile,
        'http://schemas.xmlsoap.org/ws/2004/08/eventing': EventingSchemaFile,
        'http://schemas.xmlsoap.org/ws/2004/09/mex': MetaDataExchangeSchemaFile,
        'http://docs.oasis-open.org/ws-dd/ns/dpws/2009/01': DPWSSchemaFile,
        'http://docs.oasis-open.org/ws-dd/discovery/1.1/os/wsdd-discovery-1.1-schema-os.xsd': WSDiscoverySchemaFile,
        'http://schemas.xmlsoap.org/wsdl/': WSDLSchemaFile
    }


class SDC_v1_Definitions(BaseDefinitions):
    BICEPSNamespace_base = 'http://standards.ieee.org/downloads/11073/11073-10207-2017/'
    BICEPSNamespace = 'http://standards.ieee.org/downloads/11073/11073-10207-2017/SP'
    DPWS_SDCNamespace = _DPWS_SDCNamespace  #'http://standards.ieee.org/downloads/11073/11073-20701-2018'
    MedicalDeviceTypeNamespace = 'http://standards.ieee.org/downloads/11073/11073-20702-2016'
    MessageModelNamespace = 'http://standards.ieee.org/downloads/11073/11073-10207-2017/message'
    ParticipantModelNamespace = 'http://standards.ieee.org/downloads/11073/11073-10207-2017/participant'
    ExtensionPointNamespace = 'http://standards.ieee.org/downloads/11073/11073-10207-2017/extension'
    MDPWSNameSpace = 'http://standards.ieee.org/downloads/11073/11073-20702-2016' # this name changed between WPF and SDC! IEEE correction
    MedicalDeviceType = etree_.QName(MedicalDeviceTypeNamespace, 'MedicalDevice')
    SDCDeviceType = etree_.QName(DPWS_SDCNamespace, 'SdcDevice')
    MessageModelSchemaFile = os.path.join(schemaFolder, 'BICEPS_MessageModel.xsd')
    ParticipantModelSchemaFile = os.path.join(schemaFolder, 'BICEPS_ParticipantModel.xsd')
    ExtensionPointSchemaFile = os.path.join(schemaFolder, 'ExtensionPoint.xsd')
    ActionsNamespace = DPWS_SDCNamespace
    PortTypeNamespace = DPWS_SDCNamespace

    MedicalDeviceTypesFilter = [BaseDefinitions.DpwsDeviceType, MedicalDeviceType]
    sc = sc_final
    dc = dc_final
    Actions = _SDC_v1_Actions
    SchemaFilePaths = SchemaPathsSdc
SDC_v1_Definitions.xml_validator = mk_schema_validator(SchemaResolver(SDC_v1_Definitions))



from .. import namespaces
from .. import sdcdevice
from .. import pmtypes
from ..nomenclature import NomenclatureCodes as nc
from . import providerbase

# coded values for SDC ntp and time zone
MDC_OP_SET_TIME_SYNC_REF_SRC = pmtypes.CodedValue(nc.MDC_OP_SET_TIME_SYNC_REF_SRC)
MDC_ACT_SET_TIME_ZONE = pmtypes.CodedValue(nc.MDC_ACT_SET_TIME_ZONE)

OP_SET_NTP = pmtypes.CodedValue(nc.OP_SET_NTP)
OP_SET_TZ = pmtypes.CodedValue(nc.OP_SET_TZ)

class GenericSDCClockProvider(providerbase.ProviderRole):
    """ Handles operations for setting ntp server and time zone.
    It guarantees that mdib has a clock descriptor and that there operations for setting
    ReferenceSource and Timezone of clock state."""

    def __init__(self, log_prefix):
        super(GenericSDCClockProvider, self).__init__(log_prefix)
        self._set_ntp_operations = []
        self._set_tz_operations = []

    def initOperations(self, mdib):
        super(GenericSDCClockProvider, self).initOperations(mdib)
        # create a clock descriptor and state if they do not exist in mdib
        clockDescriptor = self._mdib.descriptions.NODETYPE.getOne(namespaces.domTag('ClockDescriptor'), allowNone=True)
        if clockDescriptor is None:
            mdsContainer = self._mdib.descriptions.NODETYPE.getOne(namespaces.domTag('MdsDescriptor'))
            clock_descr_handle = 'clock_' + mdsContainer.handle
            self._logger.info('creating a clock descriptor, handle={}'.format(clock_descr_handle))
            clockDescriptor = self._mdib.createClockDescriptorContainer(handle=clock_descr_handle,
                                                                        parentHandle=mdsContainer.handle,
                                                                        codedValue=pmtypes.CodedValue(123),
                                                                        safetyClassification=pmtypes.SafetyClassification.INF)

        clockState = self._mdib.states.descriptorHandle.getOne(clockDescriptor.handle, allowNone = True)
        if clockState is None:
            clockState = self._mdib.mkStateContainerFromDescriptor(clockDescriptor)
            self._mdib.addState(clockState)


    def makeOperationInstance(self, operationDescriptorContainer):
        if operationDescriptorContainer.coding in (MDC_OP_SET_TIME_SYNC_REF_SRC.coding, OP_SET_NTP.coding):
            self._logger.info('instantiating "set ntp server" operation from existing descriptor handle={}'.format(
                operationDescriptorContainer.handle))
            set_ntp_operation = self._mkOperationFromOperationDescriptor(operationDescriptorContainer,
                                                                         currentArgumentHandler=self._setNTPString)
            self._set_ntp_operations.append(set_ntp_operation)
            return set_ntp_operation
        elif operationDescriptorContainer.coding in (MDC_ACT_SET_TIME_ZONE.coding, OP_SET_TZ.coding):
            self._logger.info('instantiating "set time zone" operation from existing descriptor handle={}'.format(
                operationDescriptorContainer.handle))
            set_tz_operation = self._mkOperationFromOperationDescriptor(operationDescriptorContainer,
                                                                        currentArgumentHandler=self._setTZString)
            self._set_tz_operations.append(set_tz_operation)
            return set_tz_operation
        return None  # ?

    def makeMissingOperations(self):
        ops = []
        mdsContainer = self._mdib.descriptions.NODETYPE.getOne(namespaces.domTag('MdsDescriptor'))
        clockDescriptor = self._mdib.descriptions.NODETYPE.getOne(namespaces.domTag('ClockDescriptor'), allowNone=True)
        if not self._set_ntp_operations:
            self._logger.info('adding "set ntp server" operation, code = {}'.format(nc.MDC_OP_SET_TIME_SYNC_REF_SRC))
            set_ntp_operation = self._mkOperation(sdcdevice.sco.SetStringOperation,
                                                 handle='SET_NTP_SRV_'+ mdsContainer.handle,
                                                 operationTargetHandle=clockDescriptor.handle,
                                                 codedValue=MDC_OP_SET_TIME_SYNC_REF_SRC,
                                                 currentArgumentHandler=self._setNTPString)
            self._set_ntp_operations.append(set_ntp_operation)
            ops.append(set_ntp_operation)
        if not self._set_tz_operations:
            self._logger.info('adding "set time zone" operation, code = {}'.format(nc.MDC_ACT_SET_TIME_ZONE))
            set_tz_operation = self._mkOperation(sdcdevice.sco.SetStringOperation,
                                                 handle='SET_TZONE_'+ mdsContainer.handle,
                                                 operationTargetHandle=clockDescriptor.handle,
                                                 codedValue=MDC_ACT_SET_TIME_ZONE,
                                                 currentArgumentHandler=self._setTZString)
            self._set_tz_operations.append(set_tz_operation)
            ops.append(set_tz_operation)
        return ops


    def _setNTPString(self, operationInstance, value):
        """This is the handler for the set ntp server operation.
         It sets the ReferenceSource value of clock state"""
        operationDescriptorHandle = operationInstance.handle
        operationDescriptorContainer = self._mdib.descriptions.handle.getOne(operationDescriptorHandle)
        operationTargetHandle = operationDescriptorContainer.OperationTarget
        self._logger.info('set value {} from {} to {}', operationTargetHandle, operationInstance.currentValue,
                          value)
        with self._mdib.mdibUpdateTransaction() as mgr:
            state = mgr.getComponentState(operationTargetHandle)
            if state.NODETYPE == namespaces.domTag('MdsState'):
                mdsHandle = state.descriptorHandle
                mgr.ungetState(state)
                # look for the ClockState child
                clockDescriptors = self._mdib.descriptions.NODETYPE.get(namespaces.domTag('ClockDescriptor'),[])
                clockDescriptors = [ c for c in clockDescriptors if c.parentHandle == mdsHandle]
                if len(clockDescriptors) == 1:
                    state = mgr.getComponentState(clockDescriptors[0].handle)
            if state.NODETYPE != namespaces.domTag('ClockState'):
                raise RuntimeError('_setNTPString: expected ClockState, got {}'.format(state.NODETYPE.localname))
            state.ReferenceSource = [pmtypes.ElementWithTextOnly(value)]


    def _setTZString(self, operationInstance, value):
        """This is the handler for the set time zone operation.
         It sets the TimeZone value of clock state."""
        operationDescriptorHandle = operationInstance.handle
        operationDescriptorContainer = self._mdib.descriptions.handle.getOne(operationDescriptorHandle)
        operationTargetHandle = operationDescriptorContainer.OperationTarget
        self._logger.info('set value {} from {} to {}', operationTargetHandle, operationInstance.currentValue,
                          value)
        with self._mdib.mdibUpdateTransaction() as mgr:
            state = mgr.getComponentState(operationTargetHandle)
            if state.NODETYPE == namespaces.domTag('MdsState'):
                mdsHandle = state.descriptorHandle
                mgr.ungetState(state)
                # look for the ClockState child
                clockDescriptors = self._mdib.descriptions.NODETYPE.get(namespaces.domTag('ClockDescriptor'),[])
                clockDescriptors = [ c for c in clockDescriptors if c.parentHandle == mdsHandle]
                if len(clockDescriptors) == 1:
                    state = mgr.getComponentState(clockDescriptors[0].handle)

            if state.NODETYPE != namespaces.domTag('ClockState'):
                raise RuntimeError('_setNTPString: expected Clockstate, got {}'.format(state.NODETYPE.localname))
            state.TimeZone = value



from ..pysoap.soapenvelope import AdressingFault
from ..pysoap.soapenvelope import MODE_PUSH
from ..pysoap.soapenvelope import SoapFault
from ..pysoap.soapenvelope import SoapFaultCode


class HTTPRequestHandlingError(Exception):
    """ This class is used to communicate errors from http request handlers back to http server."""

    def __init__(self, status, reason, soapfault):
        """
        :param status: integer, e.g. 404
        param reason: the provided human readable text
        """
        super(HTTPRequestHandlingError, self).__init__()
        self.status = status
        self.reason = reason
        self.soapfault = soapfault

    def __repr__(self):
        if self.soapfault:
            return '{}(status={}, reason={}'.format(self.__class__.__name__, self.status, self.soapfault)
        else:
            return '{}(status={}, reason={}'.format(self.__class__.__name__, self.status, self.reason)


class FunctionNotImplementedError(HTTPRequestHandlingError):
    def __init__(self, request):
        fault = SoapFault(request, code=SoapFaultCode.RECEIVER, reason='not implemented', details='sorry!')
        super().__init__(500, 'not implemented', fault.as_xml())


class InvalidActionError(HTTPRequestHandlingError):
    def __init__(self, request):
        fault = AdressingFault(request,
                               code=SoapFaultCode.SENDER,
                               reason='invalid action {}'.format(request.address.action))
        super().__init__(400, 'Bad Request', fault.as_xml())


class InvalidPathError(HTTPRequestHandlingError):
    def __init__(self, request, path):
        fault = AdressingFault(request,
                               code=SoapFaultCode.SENDER,
                               reason='invalid path {}'.format(path))
        super().__init__(400, 'Bad Request', fault.as_xml())


class InvalidMessageError(HTTPRequestHandlingError):
    def __init__(self, request, detail):
        fault = AdressingFault(request,
                               code=SoapFaultCode.SENDER,
                               reason='The message is not valid and cannot be processed.',
                               details='Detail: {} - The invalid message: {}'.format(detail,
                                                                                     request.rawdata.decode("utf-8")))
        super().__init__(400, 'Bad Request', fault.as_xml())


class DeliveryModeRequestedUnavailableError(HTTPRequestHandlingError):
    def __init__(self, request, detail=None):
        if detail is None:
            detail = f"The only supported mode: {MODE_PUSH}"
        fault = AdressingFault(request,
                               code=SoapFaultCode.SENDER,
                               reason='The requested delivery mode is not supported.',
                               details=detail)
        super().__init__(400, 'Bad Request', fault.as_xml())



"""Compression module for pysdc."""
from collections import OrderedDict
import zlib
try:
    import lz4.frame
except ImportError:
    lz4 = None

GZIP = 'gzip'
LZ4 = 'x-lz4'
ANY = 'any'

encodings = []
if lz4 is not None:
    encodings.append(LZ4)
encodings.append(GZIP)

class CompressionException(Exception):
    pass

class CompressionHandler(object):
    """Compression handler mixin.
    Should be used by servers and clients that are supposed to handle compression
    """
    available_encodings = encodings # initial default

    @classmethod
    def compressPayload(cls, algorithm, payload):
        """Compresses payload based on required algorithm.
        Raises CompressionException if algorithm is not supported.

        :param algorithm: one of available values specified as constants in this module
        :param payload: text to compress
        @return: compressed content
        """
        if algorithm == GZIP:
            return cls._gzip_encode(payload)
        elif algorithm == LZ4 and lz4 is not None:
            return lz4.frame.compress(payload)
        else:
            if lz4 is not None:
                raise CompressionException("{} compression is not supported. "
                                           "Only gzip and lz4 are supported".format(algorithm))
            else:
                raise CompressionException("{} compression is not supported. "
                                           "Only gzip is supported".format(algorithm))

    @staticmethod
    def decompress(payload, algorithm):
        """Compresses payload based on required algorithm.
        Raises CompressionException if algorithm is not supported.

        :param algorithm: one of available values specified as constants in this module
        :param payload: text to decompress
        @return: decompressed content
        """
        if algorithm == GZIP:
            return zlib.decompress(payload, 16 + zlib.MAX_WBITS)
        elif algorithm == LZ4 and lz4 is not None:
            return lz4.frame.decompress(payload)
        else:
            if lz4 is not None:
                raise CompressionException("{} compression is not supported. "
                                           "Only gzip and lz4 are supported".format(algorithm))
            else:
                raise CompressionException("{} compression is not supported. "
                                           "Only gzip is supported".format(algorithm))

    @staticmethod
    def _gzip_encode(payload):
        gzip_compress = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION, zlib.DEFLATED, 16 + zlib.MAX_WBITS)
        data = gzip_compress.compress(payload) + gzip_compress.flush()
        return data

    @staticmethod
    def parseHeader(header):
        """
        Examples of headers are:  Examples of its use are:

       Accept-Encoding: compress, gzip
       Accept-Encoding:
       Accept-Encoding: *
       Accept-Encoding: compress;q=0.5, gzip;q=1.0
       Accept-Encoding: gzip;q=1.0, identity; q=0.5, *;q=0

        returns sorted list of compression algorithms by priority
        """
        # for now work with standard python containers
        # if performance becomes an issue could be done within one loop
        parsedHeaders = OrderedDict()
        if header:
            for alg in (x.split(";") for x in header.split(",")):
                algName = alg[0].strip()
                parsedHeaders[algName] = None
                try:
                    parsedHeaders[algName] = float(alg[1].split("=")[1])
                except:
                    parsedHeaders[algName] = 1 # default
        return [pair[0] for pair in sorted(parsedHeaders.items(), key=lambda kv: kv[1], reverse=True)]



import time
import random
import traceback
from threading import Thread, Event
from .. import namespaces
from ..pmtypes import  AlertActivation, AlertConditionKind, AlertSignalPresence, SystemSignalActivation
from ..pmtypes import  AlertSignalManifestation
from . import providerbase


class GenericAlarmProvider(providerbase.ProviderRole):
    WORKERTHREAD_INTERVAL = 1.0 # seconds
    def __init__(self, log_prefix):
        super().__init__(log_prefix)

        # some time stamps for handling of delegable alert signals
        self._lastActivateAllDelegableAlerts = 0 # time when _activateAllDelegableAlertSignals has been called last time
        self._lastSetAlertSignalState = {} # a lookup by alert signal handle , value = time of last call

        self._stopWorker = Event()
        self._workerThread = None

    def initOperations(self, mdib):
        super().initOperations(mdib)
        self._setAlertSystemStatesInitialValues()
        self._setAlertStatesInitialValues()
        self._workerThread = Thread(target=self._workerThreadLoop)
        self._workerThread.daemon = True
        self._workerThread.start()

    def makeMissingOperations(self):
        return []

    def stop(self):
        self._stopWorker.set()
        self._workerThread.join()


    def makeOperationInstance(self, operationDescriptorContainer):
        operationTargetHandle = operationDescriptorContainer.OperationTarget
        operationTargetDescr = self._mdib.descriptions.handle.getOne(operationTargetHandle) # descriptor container
        if operationDescriptorContainer.NODETYPE == namespaces.domTag('SetValueOperationDescriptor'):
            pass
        elif operationDescriptorContainer.NODETYPE == namespaces.domTag('ActivateOperationDescriptor'):
            pass
        elif operationDescriptorContainer.NODETYPE == namespaces.domTag('SetAlertStateOperationDescriptor'):
            if operationTargetDescr.NODETYPE == namespaces.domTag('AlertSignalDescriptor') and operationTargetDescr.SignalDelegationSupported:
                modifiable_data = [d.text for d in operationDescriptorContainer.ModifiableData]
                if 'Presence' in modifiable_data \
                        and 'ActivationState' in modifiable_data \
                        and 'ActualSignalGenerationDelay' in modifiable_data:
                    # ToDo:  check for appropriate code
                    operation = self._mkOperationFromOperationDescriptor(
                        operationDescriptorContainer,
                        currentArgumentHandler=self._delegate_alert_signal,
                        timeoutHandler=self._end_delegate_alert_signal)

                    self._logger.info(f'GenericAlarmProvider: added handler "self._delegate_alert_signal" '
                                      f'for {operationDescriptorContainer} target= {operationTargetDescr} ')
                    return operation

        return None  # None == no handler for this operation instantiated

    def _setAlertSystemStatesInitialValues(self):
        alertSystemStates = self._mdib.states.NODETYPE.get(namespaces.domTag('AlertSystemState'), [])
        for alertSystemState in alertSystemStates:
            if hasattr(alertSystemState, 'SystemSignalActivation'): # attribute not exists in Draft6
                alertSystemState.SystemSignalActivation.append(SystemSignalActivation(manifestation=AlertSignalManifestation.AUD,
                                                                                              state=AlertActivation.ON))

    def _setAlertStatesInitialValues(self):
        """
        - set all AlertConditions ActivationState to 'On','
        - set all local alert Signals ActivationState to 'On'
        - set all remote alert Signals ActivationState to 'Off' (must be explicitly enabled by delegating device)"""
        alert_conditions = self._mdib.states.NODETYPE.get(namespaces.domTag('AlertConditionState'),[]) \
                         + self._mdib.states.NODETYPE.get(namespaces.domTag('LimitAlertConditionState'),[])
        for alert_condition in alert_conditions:
            alert_condition.ActivationState = AlertActivation.ON
        for alert_signal_state in self._mdib.states.NODETYPE.get(namespaces.domTag('AlertSignalState'),[]):
            alert_signal_descr = self._mdib.descriptions.handle.getOne(alert_signal_state.descriptorHandle)
            if alert_signal_descr.SignalDelegationSupported:
                alert_signal_state.Location = 'Rem'
                alert_signal_state.ActivationState = AlertActivation.OFF
                alert_signal_state.Presence = AlertSignalPresence.OFF
            else:
                alert_signal_state.ActivationState = AlertActivation.ON
                alert_signal_state.Presence = AlertSignalPresence.OFF

    def _getDescriptor(self, handle, mdib, transaction):
        """ Helper that looks for descriptor first in current transaction, then in mdib. returns first found one or raises KeyError"""
        descriptor = None
        tr_item = transaction.descriptorUpdates.get(handle)
        if tr_item is not None:
            descriptor = tr_item.new
        if descriptor is None:
            # it is not part of this transaction
            descriptor = mdib.descriptions.handle.getOne(handle, allowNone=True)
        if descriptor is None:
            raise KeyError('there is no descriptor for {}'.format(handle))
        return descriptor

    def _getChangedAlertConditionStates(self, transaction):
        result = []
        for item in list(transaction.alertStateUpdates.values()):
            tmp = item.old if item.new is None else item.new
            if tmp.NODETYPE in(namespaces.domTag('AlertConditionState'),
                               namespaces.domTag('LimitAlertConditionState')):
                result.append(tmp)
        return result

    def _findAlertSystemsWithModifications(self, mdib, transaction):
        # find all alert systems with changed states
        alertSystemStates = set()
        changedAlertConditions = self._getChangedAlertConditionStates(transaction)
        for tmp in changedAlertConditions:
            alertDescriptor = self._getDescriptor(tmp.descriptorHandle, mdib, transaction)
            alertSystemDescriptor = self._getDescriptor(alertDescriptor.parentHandle, mdib, transaction)
            if alertSystemDescriptor.handle in transaction.alertStateUpdates:
                tmp_st = transaction.alertStateUpdates[alertSystemDescriptor.handle]
                if tmp_st.new is not None:
                    alertSystemStateContainer = tmp_st.new
                    alertSystemStates.add(alertSystemStateContainer)
            else:
                alertSystemStateContainer = transaction.getAlertState(alertSystemDescriptor.handle)
                alertSystemStates.add(alertSystemStateContainer)
        return alertSystemStates

    def onPreCommit(self, mdib, transaction):
        if not transaction.alertStateUpdates:
            return
        # find all alert systems with changed states
        alertSystemStates = self._findAlertSystemsWithModifications( mdib, transaction)
        if alertSystemStates:
            # add found alert system states to transaction. This is not part of cyclich self check
            self._updateAlertSystemStates(mdib, transaction, alertSystemStates, is_self_check=False)

        # change AlertSignal Settings in order to be compliant with changed Alert Conditions
        changedAlertConditions = self._getChangedAlertConditionStates(transaction)
        for changedAlertCondition in changedAlertConditions:
            self._updateAlertSignals(changedAlertCondition, mdib, transaction)

    def _updateAlertSystemStates(self, mdib, transaction, alertSystemStates, is_self_check=True):
        """
        update alert system states
        :param mdib:
        :param transaction:
        :param alertSystemStates: list of AlertSystemStateContainer instances
        :param is_self_check: if True, LastSelfCheck and SelfCheckCount are set
        :return:
        """
        def _getAlertState(descriptorHandle):
            alertState = None
            tr_item = transaction.getStateTransactionItem(descriptorHandle)
            if tr_item is not None:
                alertState = tr_item.new
            if alertState is None:
                # it is not part of this transaction
                alertState = mdib.states.descriptorHandle.getOne(descriptorHandle, allowNone=True)
            if alertState is None:
                raise RuntimeError('there is no alert state for {}'.format(descriptorHandle))
            return alertState

        for st in alertSystemStates:
            # find all alert condition descriptors of alert system
            all_child_descriptors = mdib.descriptions.parentHandle.get(st.descriptorHandle, list())
            all_child_descriptors.extend([i.new for i in transaction.descriptorUpdates.values() if i.new.parentHandle == st.descriptorHandle])
            all_alert_condition_descr = [d for d in all_child_descriptors if hasattr(d, 'Kind')]
            # select all state containers with technical alarms present
            all_tech_descr = [ d for d in all_alert_condition_descr if d.Kind == AlertConditionKind.TECHNICAL]
            all_tech_states = [_getAlertState(d.handle) for d in all_tech_descr]
            all_tech_states = [s for s in all_tech_states if s is not None]
            all_present_tech_states = [s for s in all_tech_states if s.Presence]
            # select all state containers with physiological alarms present
            all_phys_descr = [ d for d in all_alert_condition_descr if d.Kind == AlertConditionKind.PHYSIOLOGICAL]
            all_phys_states = [_getAlertState(d.handle) for d in all_phys_descr]
            all_phys_states = [ s for s in all_phys_states if s is not None]
            all_present_phys_states = [s for s in all_phys_states if s.Presence]

            st.PresentTechnicalAlarmConditions = [s.descriptorHandle for s in all_present_tech_states]
            st.PresentPhysiologicalAlarmConditions = [s.descriptorHandle for s in all_present_phys_states]
            if is_self_check:
                st.LastSelfCheck = time.time()
                st.SelfCheckCount = 1 if st.SelfCheckCount is None else st.SelfCheckCount + 1

    def _updateAlertSignals(self, changedAlertCondition, mdib, transaction):
        """ Handle alert signals for a changed alert condition.
        This method only changes states of local signals.
        Handling of delegated signals is in the responsibility of the delegated device!"""
        alertSignalDescriptors = mdib.descriptions.ConditionSignaled.get(changedAlertCondition.descriptorHandle, [])
        # separate remote from local
        remoteAlertSignalDescriptors = [a for a in alertSignalDescriptors if a.SignalDelegationSupported]
        localAlertSignalDescriptors = [a for a in alertSignalDescriptors if not a.SignalDelegationSupported]

        # look for active delegations (we only need the Manifestation value here)
        active_delegate_manifestations = []
        for a in remoteAlertSignalDescriptors:
            alertSignalState = mdib.states.descriptorHandle.getOne(a.handle)
            if alertSignalState.Presence != AlertSignalPresence.OFF and alertSignalState.Location == 'Rem':
                active_delegate_manifestations.append(a.Manifestation)

        # this lookup gives the values that a local signal shall have:
        # key = (Cond.Presence, isDelegated): value = (SignalState.ActivationState, SignalState.Presence)
        # see BICEPS standard table 9: valid combinations of alert activation states, alert condition presence, ...
        # this is the relevant subset for our case
        lookup = {(True, True): (AlertActivation.PAUSED, AlertSignalPresence.OFF),
                  (True, False): (AlertActivation.ON, AlertSignalPresence.ON),
                  (False, True): (AlertActivation.PAUSED, AlertSignalPresence.OFF),
                  (False, False): (AlertActivation.ON, AlertSignalPresence.OFF)
                  }
        for a in localAlertSignalDescriptors:
            tr_item = transaction.getStateTransactionItem(a.handle)
            if tr_item is None:
                isDelegated = a.Manifestation in active_delegate_manifestations  # is this local signal delegated?
                activation, presence = lookup[(changedAlertCondition.Presence, isDelegated)]
                alertSignalState = transaction.getAlertState(a.handle)

                if alertSignalState.ActivationState != activation or alertSignalState.Presence != presence:
                    alertSignalState.ActivationState = activation
                    alertSignalState.Presence = presence
                else:
                    # don't change
                    transaction.ungetState(alertSignalState)

    def _setUpperLimit(self, operationDescriptorContainer, value):
        """ set upper limit of an LimitAlertConditionStateContainer"""
        operationTargetHandle = operationDescriptorContainer.operationTarget
        with self._mdib.mdibUpdateTransaction() as mgr:
            operationStateContainer = mgr._deviceMdibContainer.states.descriptorHandle.getOne(operationDescriptorContainer.handle)
            self._inAllowedRange(operationStateContainer, value)
            state = mgr.getAlertState(operationTargetHandle)
            self._logger.info('set upper limit handle="{}" from {} to {}', operationTargetHandle, state.Limits.Upper, value)
            state.Limits.Upper = value

    def _setLowerLimit(self, operationDescriptorContainer, value):
        """ set lower limit of an LimitAlertConditionStateContainer"""
        operationTargetHandle = operationDescriptorContainer.operationTarget
        with self._mdib.mdibUpdateTransaction() as mgr:
            operationStateContainer = self._mdib.states.descriptorHandle.getOne(operationDescriptorContainer.handle)
            self._inAllowedRange(operationStateContainer, value)
            state = mgr.getAlertState(operationTargetHandle)
            self._logger.info('set lower limit handle={} from {} to {}', operationTargetHandle, state.Limits.Lower, value)
            state.Limits.Lower = value

    @staticmethod
    def _inAllowedRange(operationStateContainer, value):
        if operationStateContainer.AllowedRange:
            max_upper = max([i.Upper for i in operationStateContainer.AllowedRange])
            min_lower = min([i.Lower for i in operationStateContainer.AllowedRange])
            if not min_lower <= value <= max_upper:
                raise ValueError('value "{}" to be set is not in AllowedRanges {}'.format(value, operationStateContainer.AllowedRange))

    def _getAcSourceRange(self, source):
        """
        Given the source of the AlertCondition return PhysiologicalRange
        Only returns range with Lower, Upper and StepWidth
        :param source: Source metric handle
        @return: (Range) PhysiologicalRange
        """

        for handle in source:
            state = self._mdib.states.descriptorHandle.getOne(handle)
            ranges = state.PhysiologicalRange
            # find range that has Upper Lower and Step
            for rng in ranges:
                if rng.Upper is not None and rng.Lower is not None and rng.StepWidth is not None:
                    return rng
        return None

    def _generateValueWithinRange(self, sourceRange, previousValue, isLower=True):
        """ Generates a value within range different from the previous value"""
        if sourceRange is None:
            if isLower:
                return -random.random()
            else:
                return random.random()

        exp = str(sourceRange.StepWidth)[::-1].find(".")
        if exp == -1:
            exp = 0

        value = previousValue
        for cnt in range(int((sourceRange.Upper - sourceRange.Lower) // sourceRange.StepWidth)):
            if value != previousValue: break

            if isLower:
                value = round(sourceRange.Lower + (sourceRange.StepWidth*cnt), exp)
            else:
                value = round(sourceRange.Upper - (sourceRange.StepWidth*cnt), exp)

        return value

    def _activateAllDelegableAlertSignals(self, operationDescriptorContainer, value): #pylint: disable=unused-argument
        # find all delegable Alert Signals in main alert system
        # set ActivationState to "On"
        # set all corresponding local alert signals to paused
        allAlertConditionDescriptors = self._mdib.descriptions.NODETYPE.get(namespaces.domTag('AlertConditionDescriptor'), [])
        allAlertConditionDescriptors += self._mdib.descriptions.NODETYPE.get(namespaces.domTag('LimitAlertConditionDescriptor'), [])
        self._lastActivateAllDelegableAlerts = time.time()
        with self._mdib.mdibUpdateTransaction() as mgr:
            for acd in allAlertConditionDescriptors:
                signalDescriptors = self._mdib.descriptions.ConditionSignaled.get(acd.handle, [])
                for sd in signalDescriptors:
                    if sd.SignalDelegationSupported:
                        ss = mgr.getAlertState(sd.handle)
                        if ss.ActivationState == AlertActivation.OFF:
                            ss.ActivationState = AlertActivation.ON
                            self._pauseFallbackAlertSignals(sd, signalDescriptors, mgr)
                        else:
                            mgr.ungetState(ss)

    def _pauseFallbackAlertSignals(self, delegableSignalDescriptor, allSignalDescriptors, transaction):
        # look for local fallback signal (same Manifestation), and set it to paused
        for f in self._get_fallback_signals(delegableSignalDescriptor, allSignalDescriptors):
            ss_fallback = transaction.getAlertState(f.handle)
            if ss_fallback.ActivationState != AlertActivation.PAUSED:
                ss_fallback.ActivationState = AlertActivation.PAUSED
            else:
                transaction.ungetState(ss_fallback)

    def _activateFallbackAlertSignals(self, delegableSignalDescriptor, allSignalDescriptors, transaction):
        # look for local fallback signal (same Manifestation), and set it to paused
        for f in self._get_fallback_signals(delegableSignalDescriptor, allSignalDescriptors):
            ss_fallback = transaction.getAlertState(f.handle)
            if ss_fallback.ActivationState == AlertActivation.PAUSED:
                ss_fallback.ActivationState = AlertActivation.ON
            else:
                transaction.ungetState(ss_fallback)

    def _get_fallback_signals(self, delegableSignalDescriptor, allSignalDescriptors):
        if allSignalDescriptors is None:
            allSignalDescriptors = self._mdib.descriptions.ConditionSignaled.get(
                delegableSignalDescriptor.ConditionSignaled, [])
        return [tmp for tmp in allSignalDescriptors if not tmp.SignalDelegationSupported
                and tmp.Manifestation == delegableSignalDescriptor.Manifestation
                and tmp.ConditionSignaled == delegableSignalDescriptor.ConditionSignaled]


    def _delegate_alert_signal(self, operationInstance, value):
        operationTargetHandle = operationInstance.operationTarget
        self._lastSetAlertSignalState[operationTargetHandle] = time.time()
        with self._mdib.mdibUpdateTransaction() as mgr:
            state = mgr.getAlertState(operationTargetHandle)
            self._logger.info('delegate alert signal {} of {} from {} to {}',
                              operationTargetHandle, state, state.ActivationState, value.ActivationState)
            for elem in operationInstance._descriptorContainer.ModifiableData:
                name = elem.text
                tmp = getattr(value, name)
                setattr(state, name, tmp)

            descr = self._mdib.descriptions.handle.getOne(operationTargetHandle)
            if descr.SignalDelegationSupported:
                if value.ActivationState == AlertActivation.ON:
                    self._pauseFallbackAlertSignals(descr, None, mgr)
                else:
                    self._activateFallbackAlertSignals(descr, None, mgr)

    def _end_delegate_alert_signal(self, operationInstance, _):
        operationTargetHandle = operationInstance.operationTarget
        with self._mdib.mdibUpdateTransaction() as mgr:
            state = mgr.getAlertState(operationTargetHandle)
            self._logger.info('timeout alert signal delegate operation={} target={} ',
                              operationInstance.handle, operationTargetHandle)
            state.ActivationState = AlertActivation.OFF
            descr = self._mdib.descriptions.handle.getOne(operationTargetHandle)
            self._activateFallbackAlertSignals(descr, None, mgr)

    def _workerThreadLoop(self):
        # delay start of operation
        time.sleep(5)
        shall_stop = self._stopWorker.wait(timeout=self.WORKERTHREAD_INTERVAL)
        if shall_stop:
            return

        while True:
            shall_stop = self._stopWorker.wait(timeout=self.WORKERTHREAD_INTERVAL)
            if shall_stop:
                return
            self._updateAlertSystemState_CurrentAlerts()
            #self._handleDelegateTimeouts()

    def _getAlertSystemStates_needingUpdate(self):
        """

        :return: all AlertSystemStateContainers of those last
        """
        alertStatesNeedingUpdate = []
        try:
            all_alert_systems_descr = self._mdib.descriptions.NODETYPE.get(namespaces.domTag('AlertSystemDescriptor'), list())
            for alert_system_descr in all_alert_systems_descr:
                alert_system_state = self._mdib.states.descriptorHandle.getOne(alert_system_descr.handle, allowNone=True)
                selfcheck_period = alert_system_descr.SelfCheckPeriod
                if selfcheck_period is not None:
                    last_selfcheck = alert_system_state.LastSelfCheck or 0.0
                    if time.time() - last_selfcheck >= selfcheck_period:
                        alertStatesNeedingUpdate.append(alert_system_state)
        except:
            exc = traceback.format_exc()
            self._logger.error('_getAlertSystemStates_needingUpdate: {}', exc)
        return alertStatesNeedingUpdate

    def _updateAlertSystemState_CurrentAlerts(self):
        """ updates AlertSystemState present alarms list"""
        alertStatesNeedingUpdate = self._getAlertSystemStates_needingUpdate()
        if len(alertStatesNeedingUpdate) > 0:
            try:
                with self._mdib.mdibUpdateTransaction() as tr:
                    tr_states = [tr.getAlertState(s.descriptorHandle) for s in alertStatesNeedingUpdate]
                    self._updateAlertSystemStates(self._mdib, tr, tr_states)
            except:
                exc = traceback.format_exc()
                self._logger.error('_checkAlertStates: {}', exc)

    # def _handleDelegateTimeouts(self):
    #     if self._lastActivateAllDelegableAlerts:
    #         # find the minimal invocation_effective_timeout
    #         all_op_descrs = self._mdib.descriptions.NODETYPE.get(namespaces.domTag('SetAlertStateOperationDescriptor'), [])
    #         timeouts = [op.InvocationEffectiveTimeout for op in all_op_descrs]
    #         timeouts = [t for t in timeouts if t is not None]
    #         if not timeouts:
    #             return # nothing to do
    #         minimal_invocation_effective_timeout = min(timeouts)
    #         if time.time() - self._lastActivateAllDelegableAlerts > minimal_invocation_effective_timeout:
    #             # expired, set all AlertSignalState.ActivationState to 'Off'
    #             with self._mdib.mdibUpdateTransaction() as mgr:
    #                 for op in all_op_descrs:
    #                     signalDescr = self._mdib.descriptions.handle.getOne(op.OperationTarget)
    #                     allSignalDescriptors = self._mdib.descriptions.ConditionSignaled.get(signalDescr.ConditionSignaled, [])
    #                     ss = mgr.getAlertState(signalDescr.handle)
    #                     if ss.ActivationState == AlertActivation.ON:
    #                         ss.ActivationState = AlertActivation.OFF
    #                         self._activateFallbackAlertSignals(signalDescr, allSignalDescriptors, mgr)
    #                     else:
    #                         mgr.ungetState(ss)
    #             self._lastActivateAllDelegableAlerts = 0



import copy
import typing

from lxml import etree as etree_

from .definitions_base import ProtocolsRegistry


# class BicepsSchema(object):
#     def __init__(self, definition_cls):
#         """
#
#         :param base_definition_cls: a class derived from BaseDefinitions, it contains paths to xml schema files
#         """
#         self.parser = etree_.ETCompatXMLParser()
#         self._versionRef = definition_cls
#         self.parser.resolvers.add(self._versionRef.schemaResolver)
#
#         pmSchema = self._parseFile(self._versionRef.ParticipantModelSchemaFile)
#         self.pmSchema = etree_.XMLSchema(etree=pmSchema)
#
#         bmmSchema = self._parseFile(self._versionRef.MessageModelSchemaFile)
#         self.bmmSchema = etree_.XMLSchema(etree=bmmSchema)
#
#         mexSchema = self._parseFile(self._versionRef.MetaDataExchangeSchemaFile, normalized=False)
#         self.mexSchema = etree_.XMLSchema(etree=mexSchema)
#
#         evtSchema = self._parseFile(self._versionRef.EventingSchemaFile, normalized=False)
#         self.evtSchema = etree_.XMLSchema(etree=evtSchema)
#
#         s12Schema = self._parseFile(self._versionRef.SoapEnvelopeSchemaFile, normalized=False)
#         self.s12Schema = etree_.XMLSchema(etree=s12Schema)
#
#         dpwsSchema = self._parseFile(self._versionRef.DPWSSchemaFile, normalized=False)
#         self.dpwsSchema = etree_.XMLSchema(etree=dpwsSchema)
#
#     def __str__(self):
#         return '{} {}'.format(self.__class__.__name__, self._versionRef.__name__)
#
#     def _parseFile(self, path, normalized=True):
#         with open(path, 'rb') as f:
#             xml_text = f.read()
#         if normalized:
#             xml_text = self._versionRef.normalizeXMLText(xml_text)
#         return etree_.fromstring(xml_text, parser=self.parser, base_url=path)


def _shortActionString(action):
    for cls in ProtocolsRegistry.protocols:
        if cls.ActionsNamespace is not None and action.startswith(cls.ActionsNamespace):
            return '{}:{}'.format(cls.__name__, action[len(cls.ActionsNamespace):])
    return action


def shortFilterString(actions):
    """
    Helper function to make shorter action strings for logging
    :param actions: list of strings
    :return: a comma separated string of shortened names
    """
    return ', '.join([_shortActionString(a) for a in actions])


def copy_node(node: etree_._Element, method=copy.deepcopy) -> etree_._Element:
    """
    Copy and preserve complete namespace. See https://github.com/Draegerwerk/sdc11073/issues/191

    :param node: node to be copied
    :param method: method that copies an etree element
    :return: new node
    """
    # walk from target to root
    current = node
    ns_map_list: typing.List[typing.Dict[str, str]] = []  # saves all namespaces
    while current is not None:
        ns_map_list.append({k: v for k, v in current.nsmap.items() if k})  # filter for default namespace
        current = current.getparent()

    # create new instance
    root_tree = node.getroottree()
    current = method(root_tree.getroot())
    x_path_steps = root_tree.getpath(node).split('/')[1:]
    assert len(x_path_steps) == len(ns_map_list)

    # walk from root to target
    ns_map_list.reverse()
    for i, step in enumerate(x_path_steps):
        if i == 0:
            step = f'/{step}'
        current = current.xpath(step, namespaces=ns_map_list[i])[0]
    return current


def copy_node_wo_parent(node: etree_._Element, method=copy.deepcopy) -> etree_._Element:
    """
    Copy node but only keep relevant information and no parent.

    :param node: node to be copied
    :param method: method that copies an etree element
    :return: new node
    """
    new_node = etree_.Element(node.tag, attrib=node.attrib, nsmap=node.nsmap)
    new_node.text = node.text
    new_node.tail = node.tail
    new_node.extend((method(child) for child in node))
    return new_node



""" Implementation of some data types used in Message Model"""
import enum
from .pmtypes import PropertyBasedPMType
from .namespaces import msgTag
from .mdib import containerproperties as cp

# pylint: disable=invalid-name

class RetrievabilityMethod(enum.Enum):
    GET = 'Get'
    PERIODIC = 'Per'
    EPISODIC = 'Ep'
    STREAM = 'Strm'


class RetrievabilityInfo(PropertyBasedPMType):
    Method = cp.NodeAttributeProperty('Method')
    UpdatePeriod = cp.DurationAttributeProperty('UpdatePeriod', impliedPyValue=1.0)
    _props = ['Method', 'UpdatePeriod']

    def __init__(self, method: RetrievabilityMethod, update_period: [float, None] = None):
        self.Method = method.value
        self.UpdatePeriod = update_period

    @classmethod
    def fromNode(cls, node):
        obj = cls(RetrievabilityMethod.GET, None)  # any allowed value, will be overwritten in update_node
        obj.updateFromNode(node)
        return obj

    def __repr__(self):
        return f'{self.__class__.__name__} {self.Method} period={self.UpdatePeriod}'


class Retrievability(PropertyBasedPMType):
    By = cp.SubElementListProperty([msgTag('By')], cls=RetrievabilityInfo)
    _props = ['By']

    def __init__(self, retrievability_info_list=None):
        self.By = retrievability_info_list or []

    @classmethod
    def fromNode(cls, node):
        obj = cls(None)
        obj.updateFromNode(node)
        return obj



"""
This module contains the default implementation for waveform handling of the device.
The sdc device periodically calls mdib.update_all_rt_samples method, which itself calls same method
of its waveform source. It is the responsibility of this method to update the RealtimeSampleArrayStates
of the mdib.
"""
import time
from abc import ABC, abstractmethod
from .. import pmtypes


class RtSampleArray:
    """ This class contains a list of waveform values plus time stamps and annotations.
    It is used to create Waveform notifications."""
    def __init__(self, determination_time, sample_period, samples, activation_state):
        """
        :param determination_time: the time stamp of the first value in samples
        :param sample_period: the time difference between two samples
        :param samples: a list of 2-tuples (value (float or int), flag annotation_trigger)
        :param activation_state: one of pmtypes.ComponentActivation values
        """
        self.determination_time = determination_time
        self.sample_period = sample_period
        self.samples = samples
        self.activation_state = activation_state
        self.annotations = []
        self.apply_annotations = []

    def _nearest_index(self, timestamp):
        # first check if timestamp is outside the range of this sample array. Accept 0.5*sample period as tolerance.
        if self.determination_time is None:  # when deactivated, determinationTime is None
            return None
        if timestamp < (self.determination_time - self.sample_period * 0.5):
            return None
        elif timestamp >= self.determination_time + len(self.samples) * self.sample_period + self.sample_period * 0.5:
            return None
        n = (timestamp - self.determination_time) / self.sample_period
        return int(n) + 1 if n % 1 >= 0.5 else int(n)

    def get_annotation_trigger_timestamps(self):
        """ returns the time stamps of all samples annotation_trigger set"""
        return [self.determination_time + i * self.sample_period for i, sample in enumerate(self.samples) if sample[1]]

    def add_annotations_at(self, annotation, timestamps):
        """
        :param annotation: a pmtypes.Annotation instance
        :param timestamps: a list of time stamps (time.time based)
        """
        applied = False
        annotation_index = len(self.annotations)  # Index is zero-based
        for t in timestamps:
            i = self._nearest_index(t)
            if i is not None:
                self.apply_annotations.append(pmtypes.ApplyAnnotation(annotation_index, i))
                applied = True
        if applied:
            self.annotations.append(annotation)


class _SampleArrayGenerator:
    """Wraps a waveform generator and makes RtSampleArray objects"""
    def __init__(self, descriptor_handle, generator):
        self._descriptor_handle = descriptor_handle
        self._last_timestamp = None
        self._activation_state = pmtypes.ComponentActivation.ON
        self._generator = generator
        self.current_rt_sample_array = None

    def set_activation_state(self, component_activation):
        """
        :param component_activation: one of pmtypes.ComponentActivation values
        """
        self._activation_state = component_activation
        if component_activation == pmtypes.ComponentActivation.ON:
            self._last_timestamp = time.time()

    def getNextSampleArray(self):
        """ Read sample values from waveform generator and calculate determination time.
        If activation state is not 'On', no samples are returned.
        @return: RtSampleArray instance"""
        if self._activation_state != pmtypes.ComponentActivation.ON:
            self.current_rt_sample_array = RtSampleArray(None, self._generator.sampleperiod, [], self._activation_state)
        else:
            now = time.time()
            observation_time = self._last_timestamp or now
            samples_count = int((now - observation_time) / self._generator.sampleperiod)
            samples = self._generator.nextSamples(samples_count)
            self._last_timestamp = observation_time + self._generator.sampleperiod * samples_count
            self.current_rt_sample_array = RtSampleArray(
                observation_time, self._generator.sampleperiod, samples, self._activation_state)
        return self.current_rt_sample_array

    def setWfGenerator(self, generator):
        self._generator = generator


class AbstractWaveformSource(ABC):
    """The methods declared by this abstract class are used by mdib. """
    @abstractmethod
    def update_all_realtime_samples(self, mdib):
        pass

    @abstractmethod
    def register_waveform_generator(self, mdib, descriptorHandle, wfGenerator):
        pass

    @abstractmethod
    def set_activation_state(self, mdib, descriptorHandle, componentActivation):
        pass


class DefaultWaveformSource(AbstractWaveformSource):
    """ This is the basic mechanism that reads data from waveform sources and applies it to mdib
    via real time transaction.
    Method 'update_all_realtime_samples' must be called periodically."""
    def __init__(self):
        self._waveform_generators = {}
        self._annotators = {}

    def update_all_realtime_samples(self, transaction):
        """ update all realtime sample states that have a waveform generator registered.
        On transaction commit the mdib will call the corresponding send method of the sdc device."""
        for descriptor_handle in self._waveform_generators.keys():
            st = transaction.getRealTimeSampleArrayMetricState(descriptor_handle)
            self._update_rt_samples(st)
        self._add_all_annotations()

    def register_waveform_generator(self, mdib, descriptor_handle, wf_generator):
        """
        param mdib: a device mdib instance
        :param descriptor_handle: the handle of the RealtimeSampelArray that shall accept this data
        :param wf_generator: a waveforms.WaveformGenerator instance
        """
        sample_period = wf_generator.sampleperiod
        descriptor_container = mdib.descriptions.handle.getOne(descriptor_handle)
        if descriptor_container.SamplePeriod != sample_period:
            # we must inform subscribers
            with mdib.mdibUpdateTransaction() as tr:
                descr = tr.getDescriptor(descriptor_handle)
                descr.SamplePeriod = sample_period
        if descriptor_handle in self._waveform_generators:
            self._waveform_generators[descriptor_handle].setWfGenerator(wf_generator)
        else:
            self._waveform_generators[descriptor_handle] = _SampleArrayGenerator(descriptor_handle, wf_generator)

    def set_activation_state(self, mdib, descriptorHandle, componentActivation):
        """
        param mdib: a device mdib instance
        :param descriptorHandle: a handle string
        :param componentActivation: one of pmtypes.ComponentActivation values
        """
        self._waveform_generators[descriptorHandle].set_activation_state(componentActivation)
        with mdib.mdibUpdateTransaction() as tr:
            st = tr.getRealTimeSampleArrayMetricState(descriptorHandle)
            st.ActivationState = componentActivation

    def register_annotation_generator(self, annotator, triggerHandle, annotatedHandles):
        """
        :param annotator: a pmtypes.Annotation instance
        :param triggerHandle: The handle of the waveform that triggers the annotator ( trigger = start of a waveform cycle)
        :param annotatedHandles: the handles of the waveforms that shall be annotated.
        """
        self._annotators[triggerHandle] = (annotator, annotatedHandles)

    def _update_rt_samples(self, state):
        """ update waveforms state from waveform generator (if available)"""
        wf_generator = self._waveform_generators.get(state.descriptorHandle)
        if wf_generator:
            rt_sample = wf_generator.getNextSampleArray()
            samples = [s[0] for s in rt_sample.samples]  # only the values without the 'start of cycle' flags
            if state.metricValue is None:
                state.mkMetricValue()
            state.metricValue.Samples = samples
            state.metricValue.DeterminationTime = rt_sample.determination_time
            state.metricValue.Annotations = rt_sample.annotations
            state.metricValue.ApplyAnnotations = rt_sample.apply_annotations
            state.ActivationState = rt_sample.activation_state

    def _add_all_annotations(self):
        """ add annotations to all current RtSampleArrays """
        rt_sample_arrays = {handle: g.current_rt_sample_array for (handle, g) in self._waveform_generators.items()}
        for src_handle, _annotator in self._annotators.items():
            if src_handle in rt_sample_arrays:
                annotation, dest_handles = _annotator
                timestamps = rt_sample_arrays[src_handle].get_annotation_trigger_timestamps()
                if timestamps:
                    for dest_handle in dest_handles:
                        if dest_handle in rt_sample_arrays:
                            rt_sample_arrays[dest_handle].add_annotations_at(annotation, timestamps)



import logging
import threading
import traceback
import time

from threading import Lock
from collections import deque
from collections import namedtuple
from statistics import mean, stdev
import copy
from lxml import etree as etree_
from .. import observableproperties as properties
from . import mdibbase
from . import msgreader
from .. import namespaces
from .. import pmtypes
from concurrent import futures
from .. import loghelper
from .. import xmlparsing

_global_nsmap = namespaces.nsmap

PROFILING = False
if PROFILING:
    import cProfile
    import pstats
    from io import StringIO


LOG_WF_AGE_INTERVAL = 30 # how often a log message is written with mean and stdef of waveforms age
AGE_CALC_SAMPLES_COUNT = 100 # amount of data for wf mean age and stdev calculation

A_NO_LOG = 0
A_OUT_OF_RANGE = 1
A_STILL_OUT_OF_RANGE = 2
A_BACK_IN_RANGE = 3


class DeterminationTimeWarner:
    """A Helper to reduce log warnings regarding determination time."""
    ST_IN_RANGE = 0
    ST_OUT_OF_RANGE = 1
    result_lookup = {
        # (last, current) :  (action, shall_repeat)
        (ST_IN_RANGE, ST_IN_RANGE): (A_NO_LOG, False),
        (ST_IN_RANGE, ST_OUT_OF_RANGE): (A_OUT_OF_RANGE, False),
        (ST_OUT_OF_RANGE, ST_OUT_OF_RANGE): (A_STILL_OUT_OF_RANGE, True),
        (ST_OUT_OF_RANGE, ST_IN_RANGE): (A_BACK_IN_RANGE, False)
    }
    def __init__(self, repeat_period=30):
        self.repeat_period = repeat_period
        self._last_log_time = 0
        self.last_state = self.ST_IN_RANGE

    def getOutOfDeterminationTimeLogState(self, minAge, maxAge, warn_limit):
        """
        @return: one of above constants
        """
        now = time.time()
        if minAge < -warn_limit or maxAge > warn_limit:
            current_state = self.ST_OUT_OF_RANGE
        else:
            current_state = self.ST_IN_RANGE
        action, shall_repeat = self.result_lookup[(self.last_state, current_state)]
        if self.last_state  != current_state:
            # a state transition
            self.last_state = current_state
            self._last_log_time = now
            return action
        else:
            # no state transition, but might need repeated logging
            if shall_repeat and now - self._last_log_time >= self.repeat_period:
                self._last_log_time = now
                return action
            else:
                return A_NO_LOG

_AgeData = namedtuple('_AgeData', 'mean_age stdev min_age max_age')

class ClientRtBuffer(object):
    """Collects data of one real time stream."""
    def __init__(self, sample_period, max_samples):
        """
        :param sample_period: float value, in seconds. 
                              When an incoming real time sample array is split into single RtSampleContainers, this is used to calculate the individual time stamps.
                              Value can be zero if correct value is not known. In this case all Containers will have the observation time of the sample array.
        :param max_samples: integer, max. length of self.rtdata
        """
        self.rt_data = deque(maxlen=max_samples)
        self.sample_period = sample_period
        self._max_samples = max_samples
        self._logger = loghelper.getLoggerAdapter('sdc.client.mdib.rt')
        self._lock = Lock()
        self.last_sc = None  # last statecontainer that was handled
        self._age_of_data_list = deque(maxlen=AGE_CALC_SAMPLES_COUNT) # used to calculate average age of samples when received
        self._reported_min_age = None
        self._reported_max_age = None


    def mkRtSampleContainers(self, realtimeSampleArrayContainer):
        """

        :param realtimeSampleArrayContainer: a RealTimeSampleArrayMetricStateContainer instance
        :return: a list of mdibbase.RtSampleContainer
        """
        self.last_sc = realtimeSampleArrayContainer
        metricValue = realtimeSampleArrayContainer.metricValue
        if metricValue is None:
            # this can happen if metric state is not activated.
            self._logger.debug('real time sample array "{} "has no metric value, ignoring it', realtimeSampleArrayContainer.descriptorHandle)
            return []
        observationTime = metricValue.DeterminationTime
        annotations = metricValue.Annotations
        applyAnnotations = metricValue.ApplyAnnotations
        rtSampleContainers = []
        if metricValue.Samples is not None:
            for i, sample in enumerate(metricValue.Samples):
                appliedAnnotations = []
                if applyAnnotations is not None:
                    for aa in applyAnnotations:
                        if aa.SampleIndex == i:
                            # there is an annotation for this sample:
                            aIndex = aa.AnnotationIndex
                            annot = annotations[aIndex] # index is zero-based
                            appliedAnnotations.append(annot)
                t = observationTime + i * self.sample_period
                rtSampleContainers.append(mdibbase.RtSampleContainer(sample, t, metricValue.Validity, appliedAnnotations))
        return rtSampleContainers

    def addRtSampleContainers(self, sc):
        if not sc:
            return
        with self._lock:
            self.rt_data.extend(sc)
            self._age_of_data_list.append(time.time() - sc[-1].observationTime) # use time of youngest sample, this is the best value for indication of delays
        try:
            self._reported_min_age = min(self._age_of_data_list[-1], self._reported_min_age)
        except TypeError:
            self._reported_min_age = self._age_of_data_list[-1]
        try:
            self._reported_max_age = max(self._age_of_data_list[-1], self._reported_min_age)
        except TypeError:
            self._reported_max_age = self._age_of_data_list[-1]

    def readData(self):
        """ This read method consumes all data in buffer.
        @return: a list of RtSampleContainer objects"""    
        with self._lock:
            ret = copy.copy(self.rt_data)
            self.rt_data.clear()
        return ret


    def get_age_stdev(self):
        with self._lock:
            min_value, self._reported_min_age = self._reported_min_age, None
            max_value, self._reported_max_age = self._reported_max_age, None
            mean_data = 0 if len(self._age_of_data_list) == 0 else mean(self._age_of_data_list)
            std_deviation = 0 if len(self._age_of_data_list) < 2 else stdev(self._age_of_data_list)
            return _AgeData(mean_data, std_deviation, min_value or 0, max_value or 0)


MDIB_VERSION_TOO_OLD = '{}: received too old MdibVersion, current {}, received {}'
MDIB_VERSION_UNEXPECTED = '{}: received unexpect MdibVersion, expected {}, received {}'
MDIB_VERSION_NOT_ALLOWED = '{}: received same MdibVersion (only allowed after DescriptionModificationReports), ' \
                           'expected {}, received {}'
_BufferedNotification = namedtuple('_BufferedNotification', 'report handler')


class ClientMdibContainer(mdibbase.MdibContainer):
    """ This mdib is meant to be read-only.
    Only update source is a BICEPSClient."""

    DETERMINATIONTIME_WARN_LIMIT = 1.0 # in seconds
    MDIB_VERSION_CHECK_DISABLED = False # for testing purpose you can disable checking of mdib version, so that every notification is accepted.
    INITIAL_NOTIFICATION_BUFFERING = True # if False, the response for the first incoming notification is answered after the getmdib is done.
                                          # if True, first notifications are buffered and the responses are sent immediately.
    def __init__(self, sdcClient, maxRealtimeSamples=100):
        super(ClientMdibContainer, self).__init__(sdcClient.sdc_definitions)
        self._synchronizedReports = threading.Event()
        self._logger = loghelper.getLoggerAdapter('sdc.client.mdib', sdcClient.log_prefix)
        self._sdcClient = sdcClient
        self._isInitialized = False
        self.rtBuffers = {}  # key  is a handle, value is a ClientRtBuffer
        self._maxRealtimeSamples = maxRealtimeSamples
        self._last_wf_age_log = time.time()
        if PROFILING:
            self.pr = cProfile.Profile()
        
        self._contextMdibVersion = None
        self._msgReader = msgreader.MessageReader(self)
        # a buffer for notifications that are received before initial getmdib is done
        self._bufferedNotifications = list()
        self._bufferedNotificationsLock = Lock()
        self.waveform_time_warner = DeterminationTimeWarner()
        self.metric_time_warner = DeterminationTimeWarner()

        self._last_descr_modification_mdib_version = None

    def initMdib(self):
        if  self._isInitialized:
            raise RuntimeError('ClientMdibContainer is already initialized')
        # first start receiving notifications, then call getMdib.
        # Otherwise we might miss notifications.
        self._bindToObservables()
        
        getService = self._sdcClient.client('Get')
        self._logger.info('initializing mdib...')
        mdibNode = getService.getMdibNode()
        self.nsmapper.useDocPrefixes(mdibNode.nsmap)
        self._logger.info('creating description containers...')
        descriptorContainers = self._msgReader.readMdDescription(mdibNode)
        with self.descriptions._lock: #pylint: disable=protected-access
            self.descriptions.clear()
        self.addDescriptionContainers(descriptorContainers)
        self._logger.info('creating state containers...')
        self.clearStates()
        stateContainers = self._msgReader.readMdState(mdibNode)
        self.addStateContainers(stateContainers)

        mdibVersion = mdibNode.get('MdibVersion')
        sequenceId = mdibNode.get('SequenceId')
        instanceId = mdibNode.get('InstanceId')
        if mdibVersion is not None:
            self.mdibVersion = int(mdibVersion)
            self._logger.info('setting initial mdib version to {}', mdibVersion)
        else:
            self._logger.warn('found no mdib version in GetMdib response, assuming "0"')
            self.mdibVersion = 0
        self.sequenceId = sequenceId
        self._logger.info('setting initial sequence Id to {}', sequenceId)

        if instanceId is not None:
            self.instanceId = int(instanceId)
            self._logger.info('setting initial instance id  to {}', instanceId)

        # retrieve context states only if there were none in mdibNode
        if len(self.contextStates.objects) == 0:
            self._getContextStates()
        else:
            self._logger.info('found context states in GetMdib Result, will not call getContextStates')

        # process buffered notifications
        with self._bufferedNotificationsLock:
            for bufferedReport in self._bufferedNotifications:
                bufferedReport.handler(bufferedReport.report, is_buffered_report=True)
            del self._bufferedNotifications[:]
            self._isInitialized = True

        self._sdcClient._register_mdib(self) #pylint: disable=protected-access
        self._logger.info('initializing mdib done')


    def _bufferNotification(self, report, callable):
        """
        write notification to an temporary buffer, as long as mdib is not initialized
        :param report: the report
        :param callable: the mothod that shall be called later for delayed handling of report
        :return: True if buffered, False if report shall be processed immediately
        """
        if self._isInitialized:
            # no reason to buffer
            return False

        if not self.INITIAL_NOTIFICATION_BUFFERING:
            self._waitUntilInitialized(callable.__name__)
            return False

        # get lock and check if we need to write to buffer
        with self._bufferedNotificationsLock:
            if not self._isInitialized:
                self._bufferedNotifications.append(_BufferedNotification(report, callable))
                return True
            return False

    def syncContextStates(self):
        """This method requests all context states from device and deletes all local context states that are not
        available in response from Device."""
        try:
            self._logger.info('syncContextStates called')
            contextService = self._sdcClient.client('Context')
            responseNode = contextService.getContextStatesNode()
            self._logger.info('creating context state containers...')
            contextStateContainers = self._msgReader.readContextState(responseNode)
            devices_contextStateHandles = [s.Handle for s in contextStateContainers]
            with self.contextStates._lock:  # pylint: disable=protected-access
                for obj in self.contextStates.objects:
                    if obj.Handle not in devices_contextStateHandles:
                        self.contextStates.removeObjectNoLock((obj))
        except:
            self._logger.error(traceback.format_exc())


    def _getContextStates(self, handles = None):
        try:
            self._logger.debug('new Query, handles={}', handles)
            time.sleep(0.001)
            contextService = self._sdcClient.client('Context')
            self._logger.info('requesting context states...')
            responseNode = contextService.getContextStatesNode(handles)
            self._logger.info('creating context state containers...')
            contextStateContainers = self._msgReader.readContextState(responseNode)

            self._contextMdibVersion = int(responseNode.get('MdibVersion', '0'))
            self._logger.debug('_getContextStates: setting _contextMdibVersion to {}', self._contextMdibVersion)
            
            self._logger.debug('got {} context states', len(contextStateContainers))
            with self.contextStates._lock: #pylint: disable=protected-access
                for stateContainer in contextStateContainers:
                    oldStateContainers = self.contextStates.handle.get(stateContainer.Handle, [])
                    if len(oldStateContainers) == 0:
                        self.contextStates.addObjectNoLock(stateContainer)
                        self._logger.debug('new ContextState {}', stateContainer)
                    elif len(oldStateContainers) == 1:
                        oldStateContainer = oldStateContainers[0]
                        if oldStateContainer.StateVersion != stateContainer.StateVersion:
                            self._logger.debug('update {} ==> {}', oldStateContainer, stateContainer)
                            oldStateContainer.updateFromNode(stateContainer.node)
                            self.contextStates.updateObjectNoLock(oldStateContainer)
                        else:
                            old = etree_.tostring(oldStateContainer.node)
                            new = etree_.tostring(stateContainer.node)
                            if old == new:
                                self._logger.debug('no update {}', oldStateContainer.node)
                            else:
                                self._logger.error('no update but different!\n{ \n{}',
                                    lambda:etree_.tostring(oldStateContainer.node), lambda:etree_.tostring(stateContainer.node)) #pylint: disable=cell-var-from-loop 
                    else:
                        txt = ', '.join([str(x) for x in oldStateContainers])
                        self._logger.error('found {} objects: {}', len(oldStateContainers), txt)
                    
        except:
            self._logger.error(traceback.format_exc())
        finally:
            self._logger.info('_getContextStates done')


    def _bindToObservables(self):
        # observe properties of sdcClient
        if PROFILING:
            properties.bind(self._sdcClient, waveFormReport=self._onWaveformReportProfiled)
        else:
            properties.bind(self._sdcClient, waveFormReport=self._onWaveformReport)
        properties.bind(self._sdcClient, episodicMetricReport=self._onEpisodicMetricReport)
        properties.bind(self._sdcClient, episodicAlertReport=self._onEpisodicAlertReport)
        properties.bind(self._sdcClient, episodicContextReport=self._onEpisodicContextReport)
        properties.bind(self._sdcClient, episodicComponentReport=self._onEpisodicComponentReport)
        properties.bind(self._sdcClient, descriptionModificationReport=self._onDescriptionModificationReport)
        properties.bind(self._sdcClient, episodicOperationalStateReport=self._onOperationalStateReport)


    def _canAcceptMdibVersion(self, log_prefix, newMdibVersion, is_description_modification=False):
        if self.MDIB_VERSION_CHECK_DISABLED:
            return True
        if newMdibVersion is None:
            self._logger.error('{}: could not check MdibVersion!', log_prefix)
        else:
            # log deviations from expected mdib version
            if is_description_modification and newMdibVersion > self.mdibVersion:
                self._logger.debug('{}: MdibVersion received via DescriptionModification will be allowed for next '
                                   'reports, current MdibVersion {}, received MdibVersion {}',
                                   log_prefix, self.mdibVersion, newMdibVersion)
                self._last_descr_modification_mdib_version = newMdibVersion

            if newMdibVersion < self.mdibVersion:
                log_level = logging.ERROR if self._synchronizedReports.is_set() else logging.WARNING
                self._logger.log(log_level, MDIB_VERSION_TOO_OLD, log_prefix, self.mdibVersion, newMdibVersion)
            elif (newMdibVersion - self.mdibVersion) > 1:
                if self._sdcClient.all_subscribed:
                    self._logger.error(MDIB_VERSION_UNEXPECTED, log_prefix, self.mdibVersion + 1, newMdibVersion)
            # only after DescriptionModificationReports it is allowed to receive other reports
            # with the same mdib version
            elif newMdibVersion == self.mdibVersion \
                    and newMdibVersion != self._last_descr_modification_mdib_version \
                    and self._synchronizedReports.is_set():
                self._logger.error(MDIB_VERSION_NOT_ALLOWED, log_prefix, self.mdibVersion + 1, newMdibVersion)

            if newMdibVersion >= self.mdibVersion:
                self._synchronizedReports.set()
                if not is_description_modification and newMdibVersion > self.mdibVersion:
                    self._last_descr_modification_mdib_version = None
                return True
        return False


    def _update_mdib_version_group(self, reportNode):
        self.mdibVersion = int(reportNode.get('MdibVersion', '0'))

        sequenceId = reportNode.get('SequenceId')
        if sequenceId != self.sequenceId:
            self.sequenceId = sequenceId

        instance_id = reportNode.get('InstanceId')
        if instance_id != self.instanceId:
            self.instanceId = int(instance_id)


    def _waitUntilInitialized(self, log_prefix):
        showsuccesslog = False
        started = time.monotonic()
        while not self._isInitialized:
            delay = time.monotonic() - started
            if 3 >= delay > 1:
                showsuccesslog = True
                self._logger.warn('{}: _waitUntilInitialized takes long...', log_prefix)
            elif delay > 10:
                raise RuntimeError('_waitUntilInitialized failed')
            time.sleep(1)
        delay = time.monotonic() - started
        if showsuccesslog:
            self._logger.info('{}: _waitUntilInitialized took {} seconds', log_prefix, delay)

    def _onEpisodicMetricReport(self, reportNode, is_buffered_report=False):
        # copy reportNode, further processing might change report data. Avoid side effects
        reportNode = xmlparsing.copy_node_wo_parent(reportNode)
        if not is_buffered_report and self._bufferNotification(reportNode, self._onEpisodicMetricReport):
            return
        newMdibVersion = int(reportNode.get('MdibVersion', '0'))
        if not self._canAcceptMdibVersion('_onEpisodicMetricReport', newMdibVersion):
            return

        now = time.time()
        metricsByHandle = {}
        maxAge = 0
        minAge = 0
        statecontainers = self._msgReader.readEpisodicMetricReport(reportNode)
        try:
            with self.mdibLock:
                self._update_mdib_version_group(reportNode)
                for sc in statecontainers:
                    if sc.descriptorContainer is not None and sc.descriptorContainer.DescriptorVersion != sc.DescriptorVersion:
                        self._logger.warn(
                            '_onEpisodicMetricReport: metric "{}": descriptor version expect "{}", found "{}"',
                            sc.descriptorHandle, sc.DescriptorVersion, sc.descriptorContainer.DescriptorVersion)
                        sc.descriptorContainer = None
                    try:
                        oldStateContainer = self.states.descriptorHandle.getOne(sc.descriptorHandle, allowNone=True)
                    except RuntimeError  as ex:
                        self._logger.error('_onEpisodicMetricReport, getOne on states: {}', ex)
                        continue

                    if oldStateContainer is not None:
                        if self._hasNewStateUsableStateVersion(oldStateContainer, sc, 'EpisodicMetricReport', is_buffered_report):
                            oldStateContainer.updateFromOtherContainer(sc)
                            self.states.updateObject(oldStateContainer)
                            metricsByHandle[oldStateContainer.descriptorHandle] = oldStateContainer
                    else:
                        self.states.addObject(sc)
                        metricsByHandle[sc.descriptorHandle] = sc

                    if sc.metricValue is not None:
                        # BICEPS: While Validity is "Ong" or "NA", the enclosing METRIC value SHALL not possess a determined value
                        # Also ignore determination time if measurement is invalid.
                        if sc.metricValue.Validity not in [pmtypes.MeasurementValidity.INVALID,
                                                           pmtypes.MeasurementValidity.NA,
                                                           pmtypes.MeasurementValidity.MEASUREMENT_ONGOING]:
                            observationTime = sc.metricValue.DeterminationTime
                            if observationTime is None:
                                self._logger.warn(
                                    '_onEpisodicMetricReport: metric {} version {} has no DeterminationTime',
                                    sc.descriptorHandle, sc.StateVersion)
                            else:
                                age = now - observationTime
                                minAge = min(minAge, age)
                                maxAge = max(maxAge, age)
            shall_log = self.metric_time_warner.getOutOfDeterminationTimeLogState(minAge, maxAge, self.DETERMINATIONTIME_WARN_LIMIT)
            if shall_log == A_OUT_OF_RANGE:
                self._logger.warn(
                    '_onEpisodicMetricReport mdibVersion {}: age of metrics outside limit of {} sec.: max, min = {:03f}, {:03f}',
                    newMdibVersion, self.DETERMINATIONTIME_WARN_LIMIT, maxAge, minAge)
            elif shall_log == A_STILL_OUT_OF_RANGE:
                self._logger.warn(
                    '_onEpisodicMetricReport mdibVersion {}: age of metrics still outside limit of {} sec.: max, min = {:03f}, {:03f}',
                    newMdibVersion, self.DETERMINATIONTIME_WARN_LIMIT, maxAge, minAge)
            elif shall_log == A_BACK_IN_RANGE:
                self._logger.info(
                    '_onEpisodicMetricReport mdibVersion {}: age of metrics back in limit of {} sec.: max, min = {:03f}, {:03f}',
                    newMdibVersion, self.DETERMINATIONTIME_WARN_LIMIT, maxAge, minAge)
        finally:
            self._updateStateObservables(metricsByHandle.values())

    def _onEpisodicAlertReport(self, reportNode, is_buffered_report=False):
        # copy reportNode, further processing might change report data. Avoid side effects
        reportNode = xmlparsing.copy_node_wo_parent(reportNode)
        if not is_buffered_report and self._bufferNotification(reportNode, self._onEpisodicAlertReport):
            return
        newMdibVersion = int(reportNode.get('MdibVersion', '0'))
        if not self._canAcceptMdibVersion('_onEpisodicAlertReport', newMdibVersion):
            return

        alertByHandle = {}
        allAlertContainers = self._msgReader.readEpisodicAlertReport(reportNode)
        self._logger.debug('_onEpisodicAlertReport: received {} alerts', len(allAlertContainers))
        try:
            with self.mdibLock:
                self._update_mdib_version_group(reportNode)
                for sc in allAlertContainers:
                    if sc.descriptorContainer is not None and sc.descriptorContainer.DescriptorVersion != sc.DescriptorVersion:
                        self._logger.warn(
                            '_onEpisodicAlertReport: alert "{}": descriptor version expect "{}", found "{}"',
                            sc.descriptorHandle, sc.DescriptorVersion, sc.descriptorContainer.DescriptorVersion)
                        sc.descriptorContainer = None
                    try:
                        oldStateContainer = self.states.descriptorHandle.getOne(sc.descriptorHandle, allowNone=True)
                    except RuntimeError as ex:
                        self._logger.error('_onEpisodicAlertReport, getOne on states: {}', ex)
                        continue

                    if oldStateContainer is not None:
                        if self._hasNewStateUsableStateVersion(oldStateContainer, sc, 'EpisodicAlertReport', is_buffered_report):
                            oldStateContainer.updateFromOtherContainer(sc)
                            self.states.updateObject(oldStateContainer)
                            alertByHandle[oldStateContainer.descriptorHandle] = oldStateContainer
                    else:
                        self.states.addObject(sc)
                        alertByHandle[sc.descriptorHandle] = sc
        finally:
            self._updateStateObservables(alertByHandle.values())

    def _onOperationalStateReport(self, reportNode, is_buffered_report=False):
        # copy reportNode, further processing might change report data. Avoid side effects
        reportNode = xmlparsing.copy_node_wo_parent(reportNode)
        if not is_buffered_report and self._bufferNotification(reportNode, self._onOperationalStateReport):
            return
        newMdibVersion = int(reportNode.get('MdibVersion', '0'))
        if not self._canAcceptMdibVersion('_onOperationalStateReport', newMdibVersion):
            return
        operationByHandle = {}
        self._logger.info('_onOperationalStateReport: report={}', lambda:etree_.tostring(reportNode))
        allOperationStateContainers = self._msgReader.readOperationalStateReport(reportNode)
        try:
            with self.mdibLock:
                self._update_mdib_version_group(reportNode)
                for sc in allOperationStateContainers:
                    if sc.descriptorContainer is not None and sc.descriptorContainer.DescriptorVersion != sc.DescriptorVersion:
                        self._logger.warn('_onOperationalStateReport: OperationState "{}": descriptor version expect "{}", found "{}"',
                                          sc.descriptorHandle, sc.DescriptorVersion, sc.descriptorContainer.DescriptorVersion)
                        sc.descriptorContainer = None
                    try:
                        oldStateContainer = self.states.descriptorHandle.getOne(sc.descriptorHandle, allowNone=True)
                    except RuntimeError as ex:
                        self._logger.error('_onOperationalStateReport, getOne on states: {}', ex)
                        continue

                    if oldStateContainer is not None:
                        if self._hasNewStateUsableStateVersion(oldStateContainer, sc, 'OperationalStateReport', is_buffered_report):
                            oldStateContainer.updateFromOtherContainer(sc)
                            self.states.updateObject(oldStateContainer)
                            operationByHandle[oldStateContainer.descriptorHandle] = oldStateContainer
                    else:
                        self.states.addObject(sc)
                        operationByHandle[sc.descriptorHandle] = sc
        finally:
            self._updateStateObservables(operationByHandle.values())

    def _onWaveformReportProfiled(self, reportNode):
        self.pr.enable()
        self._onWaveformReport(reportNode)
        self.pr.disable()
        s = StringIO()
        ps = pstats.Stats(self.pr, stream=s).sort_stats('cumulative')
        ps.print_stats(30)
        print (s.getvalue())
        print ('total number of states: {}'.format(len(self.states._objects))) #pylint:disable=protected-access
        print ('total number of objIds: {}'.format(len(self.states._objectIDs))) #pylint:disable=protected-access
        for name, l in self.states._objectIDs.items(): #pylint:disable=protected-access
            if len(l) > 50:
                print ('object {} has {} idx references, {}'.format(name, len(l), l))


    def _onWaveformReport(self, reportNode, is_buffered_report=False):
        # copy reportNode, further processing might change report data. Avoid side effects
        reportNode = xmlparsing.copy_node_wo_parent(reportNode)
        #pylint:disable=too-many-locals
        # reportNode contains a list of msg:State nodes
        if not is_buffered_report and self._bufferNotification(reportNode, self._onWaveformReport):
            return
        newMdibVersion = int(reportNode.get('MdibVersion', '0'))
        if not self._canAcceptMdibVersion('_onWaveformReport', newMdibVersion):
            return
        waveformByHandle = {}
        waveformAge = {} # collect age of all waveforms in this report, and make one report if age is above warn limit (instead of multiple)
        allRtSampleArrayContainers = self._msgReader.readWaveformReport(reportNode)
        self._logger.debug('_onWaveformReport: {} waveforms received', len(allRtSampleArrayContainers))
        try:
            with self.mdibLock:
                self._update_mdib_version_group(reportNode)
                for new_sac in allRtSampleArrayContainers:
                    d_handle = new_sac.descriptorHandle
                    descriptorContainer = new_sac.descriptorContainer
                    if descriptorContainer is None:
                        self._logger.warn('_onWaveformReport: No Descriptor found for handle "{}"', d_handle)

                    oldStateContainer = self.states.descriptorHandle.getOne(d_handle, allowNone=True)
                    if oldStateContainer is not None:
                        if self._hasNewStateUsableStateVersion(oldStateContainer, new_sac, 'WaveformReport', is_buffered_report):
                            oldStateContainer.updateFromOtherContainer(new_sac)
                            self.states.updateObject(oldStateContainer)
                            waveformByHandle[oldStateContainer.descriptorHandle] = oldStateContainer
                    else:
                        self.states.addObject(new_sac)
                        waveformByHandle[new_sac.descriptorHandle] = new_sac

                    # add to Waveform Buffer
                    rtBuffer = self.rtBuffers.get(d_handle)
                    if rtBuffer is None:
                        if descriptorContainer is not None:
                            # read sample period
                            try:
                                sample_period = descriptorContainer.SamplePeriod or 0
                            except AttributeError:
                                sample_period = 0  # default
                        rtBuffer = ClientRtBuffer(sample_period=sample_period, max_samples=self._maxRealtimeSamples)
                        self.rtBuffers[d_handle] = rtBuffer

                    rtSampleContainers = rtBuffer.mkRtSampleContainers(new_sac)
                    rtBuffer.addRtSampleContainers(rtSampleContainers)

                    # check age
                    if len(rtSampleContainers) > 0:
                        waveformAge[d_handle] = rtSampleContainers[-1].age

                    # check descriptor version
                    if descriptorContainer.DescriptorVersion != new_sac.DescriptorVersion:
                        self._logger.error('_onWaveformReport: descriptor {}: expect version "{}", found "{}"',
                                          d_handle, new_sac.DescriptorVersion, descriptorContainer.DescriptorVersion)

            if len(waveformAge) > 0:
                minAge = min(waveformAge.values())
                maxAge = max(waveformAge.values())
                shall_log = self.waveform_time_warner.getOutOfDeterminationTimeLogState(minAge, maxAge, self.DETERMINATIONTIME_WARN_LIMIT)
                if shall_log != A_NO_LOG:
                    tmp = ', '.join('"{}":{:.3f}sec.'.format(k, v) for k,v in waveformAge.items())
                    if shall_log == A_OUT_OF_RANGE:
                        self._logger.warn('_onWaveformReport mdibVersion {}: age of samples outside limit of {} sec.: age={}!',
                                          newMdibVersion, self.DETERMINATIONTIME_WARN_LIMIT, tmp)
                    elif shall_log == A_STILL_OUT_OF_RANGE:
                        self._logger.warn('_onWaveformReport mdibVersion {}: age of samples still outside limit of {} sec.: age={}!',
                                          newMdibVersion, self.DETERMINATIONTIME_WARN_LIMIT, tmp)
                    elif shall_log == A_BACK_IN_RANGE:
                        self._logger.info('_onWaveformReport mdibVersion {}: age of samples back in limit of {} sec.: age={}',
                                          newMdibVersion, self.DETERMINATIONTIME_WARN_LIMIT, tmp)
            if LOG_WF_AGE_INTERVAL:
                now = time.time()
                if now - self._last_wf_age_log >= LOG_WF_AGE_INTERVAL:
                    age_data = self.get_wf_age_stdev()
                    self._logger.info('waveform mean age={:.1f}ms., stdev={:.2f}ms. min={:.1f}ms., max={}',
                                      age_data.mean_age*1000., age_data.stdev*1000.,
                                      age_data.min_age*1000., age_data.max_age*1000.)
                    self._last_wf_age_log = now
        finally:
            self._updateStateObservables(waveformByHandle.values())

    def _onEpisodicContextReport(self, reportNode, is_buffered_report=False):
        # copy reportNode, further processing might change report data. Avoid side effects
        reportNode = xmlparsing.copy_node_wo_parent(reportNode)
        if not is_buffered_report and self._bufferNotification(reportNode, self._onEpisodicContextReport):
            return
        newMdibVersion = int(reportNode.get('MdibVersion', '0'))
        if not self._canAcceptMdibVersion('_onEpisodicContextReport', newMdibVersion):
            return
        contextByHandle = {}
        stateContainers = self._msgReader.readEpisodicContextReport(reportNode)
        try:
            with self.mdibLock:
                self._update_mdib_version_group(reportNode)
                for sc in stateContainers:
                    try:
                        oldStateContainer = self.contextStates.handle.getOne(sc.Handle, allowNone=True)
                    except RuntimeError as ex:
                        self._logger.error('_onEpisodicContextReport, getOne on contextStates: {}', ex)
                        continue

                    if oldStateContainer is not None:
                        if self._hasNewStateUsableStateVersion(oldStateContainer, sc, 'EpisodicContextReport', is_buffered_report):
                            self._logger.info(
                                '_onEpisodicContextReport: updated context state handle = {} Descriptor Handle={} Assoc={}, Validators={}',
                                sc.Handle, sc.descriptorHandle, sc.ContextAssociation, sc.Validator)
                            oldStateContainer.updateFromOtherContainer(sc)
                            self.contextStates.updateObject(oldStateContainer)
                            contextByHandle[oldStateContainer.Handle] = oldStateContainer
                    else:
                        self.contextStates.addObject(sc)
                        self._logger.info(
                            '_onEpisodicContextReport: new context state handle = {} Descriptor Handle={} Assoc={}, Validators={}',
                            sc.Handle, sc.descriptorHandle, sc.ContextAssociation, sc.Validator)
                        contextByHandle[sc.Handle] = sc
        finally:
            self._updateStateObservables(contextByHandle.values())

    def _onEpisodicComponentReport(self, reportNode, is_buffered_report=False):
        """The EpisodicComponentReport is sent if at least one property of at least one component state has changed 
        and SHOULD contain only the changed component states.
        Components are MDSs, VMDs, Channels. Not metrics and alarms
        """
        # copy reportNode, further processing might change report data. Avoid side effects
        reportNode = xmlparsing.copy_node_wo_parent(reportNode)
        if not is_buffered_report and self._bufferNotification(reportNode, self._onEpisodicComponentReport):
            return
        newMdibVersion = int(reportNode.get('MdibVersion', '1'))
        if not self._canAcceptMdibVersion('_onEpisodicComponentReport', newMdibVersion):
            return
        componentByHandle = {}
        statecontainers = self._msgReader.readEpisodicComponentReport(reportNode)
        try:
            with self.mdibLock:
                self._update_mdib_version_group(reportNode)
                for sc in statecontainers:
                    desc_h = sc.descriptorHandle
                    if desc_h is None:
                        self._logger.error('_onEpisodicComponentReport: missing descriptor handle in {}!',
                                           lambda: etree_.tostring(sc.node))  # pylint: disable=cell-var-from-loop
                    else:
                        try:
                            oldStateContainer = self.states.descriptorHandle.getOne(desc_h, allowNone=True)
                        except RuntimeError  as ex:
                            self._logger.error('_onEpisodicComponentReport, getOne on states: {}', ex)
                            continue

                        if oldStateContainer is not None:
                            if self._hasNewStateUsableStateVersion(oldStateContainer, sc, 'EpisodicComponentReport', is_buffered_report):
                                self._logger.info(
                                    '_onEpisodicComponentReport: updated component state, handle="{}" DescriptorVersion={}',
                                    desc_h, sc.DescriptorVersion)
                                oldStateContainer.updateFromOtherContainer(sc)
                                self.states.updateObject(oldStateContainer)
                                componentByHandle[oldStateContainer.descriptorHandle] = oldStateContainer
                        else:
                            self.states.addObject(sc)
                            self._logger.info(
                                '_onEpisodicComponentReport: new component state handle = {} DescriptorVersion={}',
                                desc_h, sc.DescriptorVersion)
                            componentByHandle[sc.descriptorHandle] = sc
        finally:
            self._updateStateObservables(componentByHandle.values())


    def _onDescriptionModificationReport(self, reportNode, is_buffered_report=False):
        """The DescriptionModificationReport is sent if at least one Descriptor has been created, updated or deleted during runtime.
        It consists of 1...n DescriptionModificationReportParts.
        """
        # copy reportNode, further processing might change report data. Avoid side effects
        reportNode = xmlparsing.copy_node_wo_parent(reportNode)
        if not is_buffered_report and self._bufferNotification(reportNode, self._onDescriptionModificationReport):
            return
        newMdibVersion = int(reportNode.get('MdibVersion', '0'))
        if not self._canAcceptMdibVersion('_onDescriptionModificationReport',
                                          newMdibVersion,
                                          is_description_modification=True):
            return

        descriptions_lookup_list = self._msgReader.readDescriptionModificationReport(reportNode)
        with self.mdibLock:
            self._update_mdib_version_group(reportNode)
            for descriptions_lookup in descriptions_lookup_list:
                newDescriptorByHandle = {}
                updatedDescriptorByHandle = {}

                # -- new --
                newDescriptorContainers, new_stateContainers = descriptions_lookup[pmtypes.DescriptionModificationTypes.CREATE]
                for dc in newDescriptorContainers:
                    self.descriptions.addObject(dc)
                    self._logger.debug('_onDescriptionModificationReport: created description "{}" (parent="{}")',
                                      dc.handle, dc.parentHandle)
                    newDescriptorByHandle[dc.handle] = dc
                for sc in new_stateContainers:
                    # determine multikey
                    if sc.isContextState:
                        multikey = self.contextStates
                    else:
                        multikey = self.states
                    multikey.addObject(sc)

                # -- deleted --
                deletedDescriptorContainers, stateContainers = descriptions_lookup[pmtypes.DescriptionModificationTypes.DELETE]
                for dc in deletedDescriptorContainers:
                    self._logger.debug('_onDescriptionModificationReport: remove descriptor "{}" (parent="{}")',
                                      dc.handle, dc.parentHandle)
                    self.rmDescriptorHandleAll(dc.handle) # handling of self.deletedDescriptorByHandle inside called method

                # -- updated --
                updatedDescriptorContainers, stateContainers = descriptions_lookup[pmtypes.DescriptionModificationTypes.UPDATE]
                for dc in updatedDescriptorContainers:
                    self._logger.info('_onDescriptionModificationReport: update descriptor "{}" (parent="{}")',
                                      dc.handle, dc.parentHandle)
                    container = self.descriptions.handle.getOne(dc.handle, allowNone=True)
                    if container is None:
                        pass
                    else:
                        container.updateDescrFromNode(dc.node)
                    updatedDescriptorByHandle[dc.handle] = dc
                    # if this is a context descriptor, delete all associated states that are not in
                    # state_containers list
                    if dc.isContextDescriptor:
                        updated_handles = set([s.Handle for s in stateContainers if s.descriptorHandle == dc.handle])
                        my_handles = set([s.Handle for s in self.contextStates.descriptorHandle.get(dc.handle, [])])
                        to_be_deleted = my_handles - updated_handles
                        for handle in to_be_deleted:
                            st = multikey.handle.getOne(handle)
                            self.contextStates.removeObjectNoLock(st)
                for sc in stateContainers:
                    # determine multikey
                    if sc.isContextState:
                        multikey = self.contextStates
                        oldstateContainer = multikey.handle.getOne(sc.Handle, allowNone=True)
                    else:
                        multikey = self.states
                        oldstateContainer = multikey.descriptorHandle.getOne(sc.descriptorHandle, allowNone=True)
                    if oldstateContainer is not None:
                        oldstateContainer.updateFromOtherContainer(sc)
                        multikey.updateObject(oldstateContainer)

                # write observables for every report part separately
                if newDescriptorByHandle:
                    self.newDescriptorByHandle = newDescriptorByHandle
                if updatedDescriptorByHandle:
                    self.updatedDescriptorByHandle = updatedDescriptorByHandle
                if new_stateContainers:
                    self._updateStateObservables(new_stateContainers)
                if stateContainers:
                    self._updateStateObservables(stateContainers)


    def _hasNewStateUsableStateVersion(self, oldStateContainer, newStateContainer, reportName, is_buffered_report):
        """
        compare state versions old vs new
        :param oldStateContainer:
        :param newStateContainer:
        :param reportName: used for logging
        :return: True if new state is ok for mdib , otherwise False
        """
        diff = int(newStateContainer.StateVersion) - int(oldStateContainer.StateVersion)
        # diff == 0 can happen if there is only a descriptor version update
        if diff == 1:  # this is the perfect version
            return True
        elif diff > 1:
            self._logger.error('{}: missed {} states for state DescriptorHandle={} ({}->{})',
                               reportName,
                               diff - 1, oldStateContainer.descriptorHandle,
                               oldStateContainer.StateVersion, newStateContainer.StateVersion)
            return True  # the new version is newer, therefore it can be added to mdib
        elif diff < 0:
            if not is_buffered_report:
                self._logger.error(
                    '{}: reduced state version for state DescriptorHandle={} ({}->{}) ',
                    reportName, oldStateContainer.descriptorHandle,
                    oldStateContainer.StateVersion, newStateContainer.StateVersion)
            return False
        else:  # diff == 0:
            diffs = oldStateContainer.diff(newStateContainer)  # compares all xml attributes
            if diffs:
                self._logger.error(
                    '{}: repeated state version {} for state {}, DescriptorHandle={}, but states have different data:{}',
                    reportName, oldStateContainer.StateVersion, oldStateContainer.__class__.__name__,
                    oldStateContainer.descriptorHandle, diffs)
            return False


    def waitMetricMatches(self, handle, matchesfunc, timeout):
        """ wait until a matching metric has been received. The matching is defined by the handle of the metric and the result of a matching function.
        If the matching function returns true, this function returns.
        :param handle: The handle string of the metric of interest.
        :param matchesfunc: a callable, argument is the current state with matching handle. Can be None, in that case every state matches
        Example:
            expected = 42
            def isMatchingValue(state):
                found = state.xpath('dom:MetricValue/@Value', namespaces=nsmap) # returns a list of values, empty if nothing matches
                if found:
                    found[0] = int(found[0])
                    return [expected] == found
        :param timeout: timeout in seconds
        @return: the matching state. In cas of a timeout it raises a TimeoutError exception.
        """ 
        fut = futures.Future()
        # define a callback function that sets value of fut
        def onMetricsByHandle(metricsByHandle):
            metric = metricsByHandle.get(handle)
            if metric is not None:
                if matchesfunc is None or matchesfunc(metric):
                    fut.set_result(metric)
        try:
            properties.bind(self, metricsByHandle = onMetricsByHandle)
            begin = time.monotonic()
            ret = fut.result(timeout)
            self._logger.debug('waitMetricMatches: got result after {:.2f} seconds', time.monotonic() - begin)
            return ret
        finally:
            properties.unbind(self, metricsByHandle = onMetricsByHandle)


    def mkProposedState(self, descriptorHandle, copyCurrentState=True, handle=None):
        """ Create a new state that can be used as proposed state in according operations.
        The new state is not part of mdib!

        :param descriptorHandle: the descriptor
        :param copyCurrentState: if True, all members of existing state will be copied to new state
        :param handle: if this is a multi state class, then this is the handle of the existing state that shall be used for copy.
        :return:
        """
        descr = self.descriptions.handle.getOne(descriptorHandle)
        new_state = self.mkStateContainerFromDescriptor(descr)
        if copyCurrentState:
            lookup = self.contextStates if new_state.isContextState else self.states
            if new_state.isMultiState:
                if handle is None:  # new state
                    return new_state
                else:
                    old_state = lookup.handle.getOne(handle)
            else:
                old_state = lookup.descriptorHandle.getOne(descriptorHandle)
            new_state.updateFromOtherContainer(old_state)
        return new_state

    def get_wf_age_stdev(self):
        means = []
        stdevs = []
        mins = []
        maxs = []
        for buf in self.rtBuffers.values():
            age_data = buf.get_age_stdev()
            means.append(age_data.mean_age)
            stdevs.append(age_data.stdev)
            mins.append(age_data.min_age)
            maxs.append(age_data.max_age)
        return _AgeData(mean(means), mean(stdevs), min(mins), max(maxs))



import os
import time
from threading import Lock

D_IN = 'in'
D_OUT = 'out'

T_UDP = 'udp'
T_UDPBroadcast = 'udpB'
T_HTTP = 'http'
T_WSDL = 'wsdl'

T_HTTP_REQ = 'http_req'
T_HTTP_RESP = 'http_resp'


class NullLogger(object):
    """This is a dummy logger that does nothing."""
    def __getattr__(self, name):
        return self.do_nothing
    
    def do_nothing(self, *args, **kwargs):
        pass


class CommLogger(object):
    """This is the logger that writes communication logs."""
    def __init__(self, log_folder, log_out=False, log_in=False, broadcastIpFilter=None):
        self._log_folder = log_folder
        self._log_out = log_out
        self._log_in = log_in
        self._broadcastIpFilter = broadcastIpFilter
        self._counter = 1
        self._ioLock = Lock()

        self._mkLogFolder(log_folder)


    def setBroadcastIpFilter(self, broadcastIpFilter):
        self._broadcastIpFilter = broadcastIpFilter

    @staticmethod        
    def _mkLogFolder(path):
        if not os.path.exists(path):
            os.makedirs(path) 


    def _mkFileName(self, ipType, direction, info):
        """
        :param ipType: "tcp" or "udp"
        :param direction: "in" or "out"
        """
        assert ipType in (T_UDP, T_UDPBroadcast, T_HTTP, T_HTTP_REQ, T_HTTP_RESP, T_WSDL)
        assert direction in (D_IN, D_OUT)
        extension = 'wsdl' if ipType == T_WSDL else 'xml'
        timestring = '{:06.3f}'.format(time.time())[-8:]
        self._counter += 1
        infotext = '-{}'.format(info) if info else ''
        return '{}-{}-{}{}.{}'.format(timestring, direction, ipType, infotext, extension)

    def _writeLog(self, ttype, direction, xml, info):
        path = os.path.join(self._log_folder, self._mkFileName(ttype, direction, info))
        with self._ioLock:
            with open (path,'wb') as f:
                f.write(xml)
    
    def logBroadCastMsgOut(self, xml, info=None):
        if self._log_out:
            self._writeLog(T_UDPBroadcast, D_OUT, xml, info)
            
    def logDiscoveryMsgOut(self, ipaddr, xml, info=None):
        if self._log_out and (not self._broadcastIpFilter or self._broadcastIpFilter == ipaddr):
            self._writeLog(T_UDP, D_OUT, xml, info)
    
    def logDiscoveryMsgIn(self, ipaddr, xml):
        if self._log_in and (not self._broadcastIpFilter or self._broadcastIpFilter == ipaddr):
            self._writeLog(T_UDP, D_IN, xml, None)

    def logSoapReqIn(self, xml, info=None):
        if self._log_in:
            self._writeLog(T_HTTP_REQ, D_IN, xml, info)

    def logSoapReqOut(self, xml, info=None):
        if self._log_out:
            self._writeLog(T_HTTP_REQ, D_OUT, xml, info)

    def logSoapRespIn(self, xml, info=None):
        if self._log_in:
            self._writeLog(T_HTTP_RESP, D_IN, xml, info)

    def logSoapRespOut(self, xml, info=None):
        if self._log_out:
            self._writeLog(T_HTTP_RESP, D_OUT, xml, info)

    def logSoapSubscrMsgIn(self, xml):
        if self._log_in:
            self._writeLog(T_HTTP, D_IN, xml, 'subscr')
        
    def logWsdl(self, wsdl ):
        if self._log_in:
            self._writeLog(T_WSDL, D_IN, wsdl, None)

defaultLogger = NullLogger()

def getCommunicationLogger():
    return defaultLogger

def setCommunicationLogger(comm_logger):
    global defaultLogger
    defaultLogger = comm_logger


from . import providerbase
from .alarmprovider import GenericAlarmProvider
from .metricprovider import GenericMetricProvider
from .audiopauseprovider import GenericSDCAudioPauseProvider
from .patientcontextprovider import GenericPatientContextProvider
from . import product
